!factorization and solution in the effective cyclic reduction PARABABDCR (fact+sol)
! with only NBLOKS a power of two
PROGRAM MAING
  USE PRECISION
  USE MPI
  USE PARABABDCR_PACKAGEG

 REAL(DP), ALLOCATABLE :: MATR( : ), VECTB( : ), &
                          MYMATR_A( : ), MYFILL_IN( :,: ), MYVECTB( : )

 INTEGER, ALLOCATABLE:: MYPERM( : ), OUT( : )

!PROC,        idle of current processor
!NPROCS,      the number of processors used
!LOGNPROCS,   logdue(NPROCS) number of communication one to one with other processors
!NFORPROC,    the number of initial blocks [A(i) B(i) C(i)] of size (k+m)x(k+m*2)
!             saved in MYMATR_A
!LOGNFORPROC, logdue(NFORPROC)
!NGSTEPS,     the number of global steps of reduction, it is logdue(NBLOKS)
!LENGTHM,     length of MYMATR_A
!LENGTHVB,    length of MYVECT_B
!LENGTHF,     length of MYFILL_IN
!LENGTHP,     length of MYPERM
 INTEGER :: k, m, NBLOKS, PROC,  NPROCS, LOGNPROCS, NFORPROC, LOGNFORPROC, NGSTEPS, &
            LENGTHM, LENGTHVB, LENGTHF, LENGTHP, &
            LI, P, MYINFO, STAT( MPI_STATUS_SIZE ), I, EMPTY
!             event1a, &
!            event1b, event2a, event2b, event3a, event3b

 INTEGER, PARAMETER :: MAXNPROCS= 2

 REAL(DP) ::          Q, ERROR, time_begin, time_begindistr, time_afterfact
 CHARACTER(LEN=20) :: FILE_NAME

 INTRINSIC CEILING
  CALL MPI_INIT( IERR )
 ! CALL MPE_INIT_LOG

!time in second from some time in the past
  time_begin= MPI_WTIME()

! determine rank PROC of current proc in domain MPI_COMM_WORLD , PROC= 0,....,NPROCS-1
  CALL MPI_COMM_RANK( MPI_COMM_WORLD, PROC, IERR )

  print *,'I''m processor ', PROC

!  event1a= MPE_Log_get_event_number()
!  event1b = MPE_Log_get_event_number()
!  event2a= MPE_Log_get_event_number()
!  event2b = MPE_Log_get_event_number()
!  event3a = MPE_Log_get_event_number()
!  event3b = MPE_Log_get_event_number()
!  ierr = MPE_Describe_state( event1a, event1b, &
!                                   "Bcast bound.cond.", "black" )
!  ierr = MPE_Describe_state( event2a, event2b, &
!                                   "Scatter MATR", "grey" )
!  ierr = MPE_Describe_state( event3a, event3b, &
!                                   "Scatter VECTB", "white" )


!input dimension from file
    FILE_NAME='INPUTSIZEG'
    OPEN (10, FILE=FILE_NAME, STATUS='OLD')
    READ(10,*) k
    READ(10,*) m
    READ(10,*) NBLOKS

!    print *,'k', k, 'm', m, 'NBLOKS', NBLOKS
!    k= 2
!    m= 3
!    NBLOKS= 32


!
! compute NPROCS, LOGNFORPROC, NGSTEPS on all processor - difference respect before
! in this manner I want save the communication of INDEXMSG
!
    Q= NBLOKS/2
    I= 1
    DO WHILE ( Q>MAXNPROCS )
      Q= CEILING( Q/2 )
      I= I+1
    ENDDO
    NPROCS= Q
    LOGNFORPROC= I
    NFORPROC= 2**LOGNFORPROC
    NGSTEPS= LOGDUE( NBLOKS )
    PRINT *,'Optimum choice is ', NPROCS,' processors'


    LI= (k+m)*(k+m*2)*(NFORPROC)

!
! determine length of MYMATR_A, MYVECTB, MYFILL_IN, MYPERM  in all processor
! it will contain the initial blocks of MATR(that are NFORPROC)
! and the block received after the LOGNPROCS) point to point communication
! and the boundary condition (2) and the last block is used for
! copy the first block and apply the last 2 by 2 factorization
!
    LOGNPROCS= LOGDUE(NPROCS)
!
!determine LENGTHM, length of MYMATR_A
!
!LI=(k+m)*(k+m*2)*(NFORPROC) is the length of data from input MATR
!(m**2)*(LOGNPROCS*2) is the length of data that come from other processors
!                     after communication
!(m**2)*3 is [BA BB 0]
    LENGTHM= LI + (m**2)*(LOGNPROCS*2+3)
!
!determine LENGTHVB, length of MYVECTB
!
!(k+m)*(NFORPROC) is the length of data from input VECTB
!m*LOGNPROCS      is the length of data that come
!                 from other processors after communication
!m*1              is the length of the rhs of the boundary conditions, that is of D
    LENGTHVB= (k+m)*(NFORPROC)+ m*(LOGNPROCS+1)
!
!determine LENGTHF,  size of the second dimension of MYFILL_IN
!
!(k*2+m)*(NFORPROC/2)  is the length of the second dimension fill-in associated
!                      to the first NFORPROC/2 reduction
!                      of the couples
!                       ( Ai Bi Ci             )
!                       (       Ai+1 Bi+1 Ci+1 )
!m*(NFORPROC/2-1)   is the length of the second domension of the fill-in associated
!                   to the NFORPROC/2-1 reduction of the couples
!                      A'i C'i
!                          A'i+1 C'i+1
!m*LOGNPROCS   is the length the second dimension of the fill-in associated to the reduction
!               with the  block rows that come from other processors
!              after communication

    LENGTHF= (k*2+m)*(NFORPROC/2) + m*(NFORPROC/2-1)+ m*LOGNPROCS
!
!determine LENGTHP, length of MYPERM
!
!((k+m)*2)*(NFORPROC/2)  is the length of the permutation vector associated
!                        to the first NFORPROC/2 reduction
!                        of the couples
!                        Ai Bi Ci
!                              Ai+1 Bi+1 Ci+1
!(m*2)*(NFORPROC/2-1)   is the length of the permutation vector associated
!                       to the NFORPROC/2-1 reduction of the couples
!                       A'i C'i
!                           A'i+1 C'i+1
!(m*2)           is the length of the permutation vector in the last
!                2 by 2 linear system
!
!m*2*LOGNPROCS is the length the permutation vector associated to the reduction
!              with the  block rows that come from other processors
!              after communication
    LENGTHP=  ((k+m)*2 + m*2 )*(NFORPROC/2) + m*2*LOGNPROCS

!allocate MYMATR_A, MYFILL_IN, MYPERM and MYVECTB of  proc K

!print *,'LENGTHM', LENGTHM, ' LENGTHVB', LENGTHVB, 'LENGTHP',&
!        LENGTHP,'LENGTHF', LENGTHF

!
!allocate local array
!
    ALLOCATE( MYMATR_A( LENGTHM ), &
              MYVECTB( LENGTHVB ), &
              MYFILL_IN( m,LENGTHF ), &
              MYPERM( LENGTHP ), &
              OUT(LOGNPROCS) )

!
!read input data : MATR, VECT_B
!
    IF ( PROC.EQ.0 ) THEN

      ALLOCATE( MATR( (k+m)*(k+m*2)*NBLOKS + (m**2)*2 ), VECTB( (k+m)*NBLOKS+m ) )

! read MATR and VECTB
      CALL INPUT_PARABABDCRG( k, m, NBLOKS, MATR, VECTB )
 !   print *,' MATR= ', MATR
 !    print *,' VECTB= ', VECTB

!save the boundary condition Ba, Bb and the rhs d, in MYMATR_A and MYVECTB, respectively
      call dcopy ( 2*(m**2), MATR( (k+m)*(k+m*2)*NBLOKS+1 ), 1, MYMATR_A(LENGTHM-(m**2)*3+1),1 )
      MYVECTB( LENGTHVB-m+1:LENGTHVB )=VECTB( (k+m)*NBLOKS+1:(k+m)*NBLOKS+m )

    ENDIF

!
! broadcast [BA BB] and D from 0 to the other processors
!
  ! time_begindistr= MPI_WTIME()
  ! print *, 'Time begin bcast of Ba Bb from 0 in proc',K, &
  !         'respect begin in sec= ', time_begindistr-time_begin
  !  ierr = MPE_Log_event( event1a, 0, "start bcast Ba Bb and d ")

    CALL MPI_BCAST( MYMATR_A(LENGTHM-(m**2)*3+1),2*(m**2), &
                    MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,IERR)

    CALL MPI_BCAST(MYVECTB(LENGTHVB-m+1),m,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,IERR)
  !  ierr = MPE_Log_event( event1b, 0, "end bcast Ba Bb and d")




!determine the length of null space in MYMATR_A before communication
    EMPTY=SPACE(PROC+1,NPROCS)


!
!Scatter MATR
!
 ! time_begindistr= MPI_WTIME()
! print *, 'Time begin Matr scatter in proc' ,K,&
 ! 'respect begin in sec= ', time_begindistr-time_begin

!  ierr = MPE_Log_event( event2a, 0, "start scatter MATR")
 ! print *,'Ciao before scatter PROC ',PROC
  CALL MPI_SCATTER( MATR(1), LI, MPI_DOUBLE_PRECISION, MYMATR_A(2*(m**2)*EMPTY+1), &
                    LI, MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,IERR )
 ! print *,'Ciao after scatter PROC ',PROC
!  ierr = MPE_Log_event( event2b, 0, "end scatter MATR")

 !  print *,'end scatter matr in proc ', PROC,'MYMATR_A', MYMATR_A


!
!Scatter VECTB
!
   LI= (k+m)*NFORPROC
 ! time_begindistr= MPI_WTIME()
 ! print *, 'Time begin VECTB scatter in proc' ,PROC, &
 ! 'respect begin in sec= ', time_begindistr-time_begin

 ! print *,'Ciao before scatter vectb PROC ',PROC
 ! ierr = MPE_Log_event( event3a, 0, "start scatter VECTB")
   CALL MPI_SCATTER( VECTB(1), LI, MPI_DOUBLE_PRECISION, MYVECTB(m*EMPTY+1), &
                    LI, MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,IERR )
 ! ierr = MPE_Log_event( event3b, 0, "end scatter VECTB")
 ! print *,'Ciao after scatter vectb PROC ',PROC



! factorization
!  print *,'Hi there before calling _FACT, I''m proc', K
! in this _FACT  processor send their blocks factorized and their
! fill_in and their permutation matrix to proc 0
  time_begindistr= MPI_WTIME()

!  print *, 'Time before fact in proc',PROC, &
!           'respect begin in sec= ', time_begindistr-time_begin
  CALL PARABABDCR_FACTG( PROC, LENGTHM, LENGTHP, LENGTHF, NPROCS, &
                         LOGNPROCS, NFORPROC, LOGNFORPROC, &
                         k, m, NBLOKS, MYMATR_A(1), MYPERM(1), &
                         MYFILL_IN(1,1), MYINFO, OUT(1) )
  time_afterfact= MPI_WTIME()
 !  print *, 'Time after fact in proc',PROC, &
 !          'respect begin in sec= ', time_afterfact-time_begin

     print *,'time of the factorization in ',PROC,' is', time_afterfact-time_begindistr



  time_afterfact= MPI_WTIME()
  !print *, 'Time before sol in proc',K, &
  !         'respect begin in sec= ', time_afterfact-time_begin

  CALL PARABABDCR_SOLVG( PROC, LENGTHM, LENGTHP, LENGTHF, LENGTHVB, NPROCS, &
                         LOGNPROCS, NFORPROC, LOGNFORPROC, &
                         k, m, NBLOKS, MYMATR_A(1), MYPERM(1), &
                         MYFILL_IN(1,1), OUT(1), MYVECTB(1) )
  time_total= MPI_WTIME()
!  print *, 'Time total in proc',PROC, &
!           'respect begin, is = ', time_total-time_begin
  print *, 'time of the solution in ',PROC,' is', time_total-time_afterfact
  print *, 'time of the fact + sol in ',PROC,' is', time_total-time_begindistr

!! control of solution
!    IF (INFO.EQ.0) THEN
       ERROR= MAXVAL( ABS(1- MYVECTB ) )
!      WRITE(*,*)    'NRWBLK= ',NRWBLK,'  NBLOKS= ',NBLOKS,&
!                 'NPROCS= ',NPROCS

!    WRITE(*,*) 'proc', PROC,' solution is', MYVECTB

    WRITE(*,*) 'MAX ABS ERROR IN PROCESSOR ',PROC,' is', ERROR
!    print *, 'xteo=',XTEO
!    ENDIF
!    DEALLOCATE(A, XTEO, IPIV)
!  ENDIF
!!
  DEALLOCATE( MYMATR_A, MYVECTB, &
              MYFILL_IN, MYPERM, OUT )
  IF (PROC.EQ.0) THEN
    DEALLOCATE( MATR,VECTB )
  ENDIF
!  CALL MPE_FINISH_LOG('main4')
  CALL MPI_FINALIZE(IERR)
  STOP

  CONTAINS


!subroutine input_parababdcrg
! this routine reads MATRA and B from input files
! or using intrinsic random fortran procedure
SUBROUTINE INPUT_PARABABDCRG( k,m,NBLOKS,MATRA,VECTB )



!!!!!!!!!!Input variables!!!!!!!!!!!!
!  k,m  determine size of each block, Ai ,Bi,Ci, where
!       Ai is (k+m)xm, Bi is (k+m)xk, Ci is (k+m)xk
!  NBLOKS number of block rows [Ai Bi Ci]
!!!!!!!!!Input files!!!!!!!!!!!!!!!!!
!!!!!!!!!Output variables!!!!!!!!!!!!
!  VECTB the rhs [f;d]
!  MATRA=[A1,B1,C1,A2,B2,C2,....,Am,Bm,Cm,Ba,Bb]
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!  file INPUTSR  contains STRUCTA
!  file INPUTRHS contains the  array B=[f;d]
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! USE PRECISION
! USE PARABABDCR_PACKAGE

 !IMPLICIT NONE
 !INTEGER, PARAMETER :: DP=KIND(1.0D0)


 INTEGER,  INTENT(IN) :: k, m, NBLOKS
 REAL(DP),  INTENT(IN OUT):: MATRA( (k+m)*(k+m*2)*(NBLOKS)+(m**2)*2 )
 REAL(DP),  INTENT(OUT)   :: VECTB( (k+m)*NBLOKS+m )
! local variables
 INTEGER :: I
 CHARACTER :: ANS
 CHARACTER(LEN=20) ::  FILE_NAME


     ANS='0'


 !  choose random_numbers or input_files
     DO WHILE ((ANS.NE.'r').AND.(ANS.NE.'f').AND.(ANS.NE.'R').AND.(ANS.NE.'F'))
!       PRINT *, 'Input matrix : random or from file? (r/f)'
!       READ *, ANS
        ANS='f'
     ENDDO

     SELECT CASE (ANS)
       CASE('r', 'R')
         DO I= 1,(k+m)*(k+m*2)*NBLOKS + (m**2)*2
             CALL RANDOM_NUMBER( MATRA(I ))
             MATRA(I)= ( MATRA(I)-0.5 )*20
         ENDDO

       CASE('f', 'F')
!         PRINT *,'Insert the file name that contans the input matrix'
!         READ  *, FILE_NAME
         FILE_NAME='INPUTMATR'
         OPEN (10, FILE=FILE_NAME, STATUS='OLD')
         DO I= 1,(k+m)*(k+m*2)*NBLOKS + (m**2)*2
           READ(10,*) MATRA(I)
         ENDDO
        CLOSE(10)

     END SELECT
     ANS='0'
     DO WHILE ((ANS.NE.'r').AND.(ANS.NE.'f').AND.(ANS.NE.'R').AND.(ANS.NE.'F') &
               .AND.(ANS.NE.'S').AND.(ANS.NE.'s') )
 !      PRINT *, 'Input rhs of size ',(k+m)*NBLOKS+m,' : random or from file? (r/f)'
 !      READ *, ANS
        ANS='f'
     ENDDO

     SELECT CASE (ANS)
       CASE('r', 'R')
         DO I= 1,(k+m)*NBLOKS+m
           CALL RANDOM_NUMBER(VECTB(I))
         ENDDO


       CASE('f', 'F')
!         PRINT *,'Insert the file name that contains the input rhs'
!         READ  *, FILE_NAME
         FILE_NAME='INPUTRHS'
         OPEN (20, FILE=FILE_NAME, STATUS='OLD')
         DO I= 1,(k+m)*NBLOKS+m
           READ(20,*) VECTB(I)
         ENDDO
         CLOSE(20)
     END SELECT


  RETURN
END SUBROUTINE INPUT_PARABABDCRG



END PROGRAM MAING
