 !Paramod4.f90  - The Effective Cyclic Reduction Module
!%%%%%%%%%%%%%%
! Compiling command
!     mpif90 -c -u -mpilog  paramod4.f90
!this create the object file 
!    paramod4.o
!%%%%%%%%%%%%%
MODULE PRECISION
  INTEGER, PARAMETER :: DP=KIND(1.0D0)
END MODULE PRECISION

MODULE PARABABDCR_PACKAGE4
  USE PRECISION
  USE MPI

  CONTAINS

   SUBROUTINE PARABABDCR_FACT4( K, LENGTHK, LENGTHVBK, NPROCS, LOGNPROCS,&
                                NFORPROC, NRWBLK, NBLOKS, MYMATR_A, &
                                MYPERM, MYFILL_IN, MYINFO, OUT )
!
! Parameters
! ==========
!
! Input variables
!
!   K  integer,                  rank of processor in current communication domain
!                                K is in the set 0,...,NPROCS-1
!
!   LENGTHK integer,             size of third dimension of MYMATR_A, LENGTHK>0
!
!   NPROCS    integer,           number of processors, NPROCS>0
!
!   LOGNPROCS integer,           log2 of NPROCS
!
!   NFORPROC integer,            log2 of number of blocks [Si,Ri] in each proc
!                                in the last proc there is at most 2^NFORPROC + 1 blocks
!                                and at least 1 block, NFORPROC>0
!
!   NRWBLK integer,              the dimension of blocks S, R, Ba, Bb, NRWBLK>0
!
!   NBLOKS integer,              the number of blocks S(i) and R(i)
!                                in the babd matrix, NBLOKS>0
!
!   MYMATR_A double precision,   NRWBLK by NRWBLK by LENGTHK array,
!                                the blocks of babd matrix which are saved
!                                in current processor
!
! Output variables
!   MYMATR_A double precision,   NRWBLK by NRWBLK by LENGTHK array,
!                                part of the babd matrix factorization
!
!   MYPERM   integer,            if K>0,
!                                 NRWBLK*2 by
!                                 (LENGTHK-2)/2= (LENGTHK-2 - LOGTOINT(K.EQ.0) )/2 array,
!                                 (since LOGTOINT(.TRUE.)=1, LOGTOINT(.FALSE.)=0)
!                                 the permutations associated to the
!                                 LU factorizations computed on proc K
!                                if K=0,
!                                 NRWBLK*2 by
!                                 (LENGTHK-3)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0))/2 array,
!                                 the permutations associated to the
!                                 LU factorizations computed on proc 0, even
!                                 associated to the factorization of the last
!                                 2 by 2 block system
!
!   MYFILL_IN double precision,  if K>0, NRWBLK by NRWBLK by
!                                 (LENGTHK-2)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
!                                 the fill-in matrix generated in the reduction
!                                 routine REDUCE_BLOCK called by proc K
!                                if K=0, NRWBLK by NRWBLK by
!                                 (LENGTHK-5)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
!                                 the fill-in matrix generated in the reduction
!                                 routine REDUCE_BLOCK called by proc 0
!
!   MYINFO   integer,
!          = 0:  successful exit
!          > 0:  if MYINFO = I, the I-th computed factorization
!                (in proc K) is performed on a
!                singular matrix; therefore the babd matrix is singular
!
!
!  OUT  integer,      LOGNPROCS vector,
!         OUT(I) = 0 : processor K communicates to one processor with idle greater than K,
!                       at the I-th communication in the matrix reduction phase
!         OUT(I) = 1 : processor K communicates to one processor with idle lesser than K,
!                       at the I-th communication in the matrix reduction phase
!


    INTEGER, INTENT(IN) :: K, LENGTHK, LENGTHVBK, NPROCS, LOGNPROCS, NFORPROC, &
                           NRWBLK, NBLOKS

!dimension of MYMATR_A, MYFILL_IN, MYPERM changes through different processors
    REAL(DP), INTENT(IN OUT) :: MYMATR_A( NRWBLK, NRWBLK, LENGTHK )

    REAL(DP), INTENT(OUT)    :: MYFILL_IN( NRWBLK, NRWBLK, &
                                         ( LENGTHK-5)/2 )

    INTEGER,  INTENT(OUT)    :: MYPERM( NRWBLK*2, (LENGTHK-3)/2 ), &
                                MYINFO, OUT( LOGNPROCS )

! Local variables in each processor
    INTEGER ::  MYHSTART, MYHEND, MYINDEXP, Z, I, MYLE, L, SOURCE, TAG, &
                STATUS(MPI_STATUS_SIZE), EMERGENCYTAG, EMERGENCYMSG(2), &
                IERR, JUMP, S, P, GSTEP, MSGTYPE, JUMP1, LSTART, EMPTY, &
                PROD, event1a,event1b, & 
!event2a,event2b,event3a,event3b,event4a,event4b, event5a,event5b&
                event6a,event6b, MPE_Log_get_event_number, &
                MPE_describe_state
    REAL(DP) :: SENDARRAY(NRWBLK,NRWBLK,2)

! Lapack routine : DGETRF
! Used subroutines: REASSEMBLE, REDUCE_BLOCK


!     call MPE_INIT_LOG()

      event1a = MPE_Log_get_event_number()
      event1b = MPE_Log_get_event_number()
!      event2a = MPE_Log_get_event_number()
!      event2b = MPE_Log_get_event_number()
!      event3a = MPE_Log_get_event_number()
!      event3b = MPE_Log_get_event_number()
!      event4a = MPE_Log_get_event_number()
!      event4b = MPE_Log_get_event_number()
!      event5a = MPE_Log_get_event_number()
!      event5b = MPE_Log_get_event_number()
      event6a = MPE_Log_get_event_number()
      event6b = MPE_Log_get_event_number()
!      event7a = MPE_Log_get_event_number()
!      event7b = MPE_Log_get_event_number()


          ierr = MPE_Describe_state( event1a, event1b, &
                                     "SendRecv MYMATR_A in reduction phase", "red" )
!          ierr = MPE_Describe_state( event2a, event2b, &
!                                     "Send EMERGMSG to proc NPROCS-1", "blue" )
!          ierr = MPE_Describe_state( event3a, event3b, &
!                                     "Receive MYMATR_A in reduction phase", "orange" )
!          ierr = MPE_Describe_state( event4a, event4b, &
!                                     "Receive  EMERGMSG ", "green" )
!          ierr = MPE_Describe_state( event5a, event5b, &
!                                    "Send MYMATR_A after receiving EMERGMSG", "cyan" )
          ierr = MPE_Describe_state( event6a, event6b, &
                                    "Compute reduction - internal cycle", "yellow" )

     
      EMPTY=SPACE(K+1,NPROCS)

!
! initialize MYINDEXP
!
      MYINDEXP= 0

!
! initialize MYLE, L
!
!  MYLE current number of blocks [Si Ri] to reduce
!   in proc 0 the blocks Ba,Bb,0 are no considered
!   then MYLE is 2**NFORPROC (or NBLOKS, if NPROCS.EQ.1) at beginning
!   in proc 0,...NPROCS-2,
!  L current index of last nonnull NRWBLK x NRWBLK block in MYMATR_A,
!   in proc 0 the blocks Ba,Bb are not considered and even initial
!   zero blocks
!
!     IF ( NPROCS.GT.1 ) THEN
!       IF ( K.LT.NPROCS-1 ) THEN
         MYLE= 2**NFORPROC
!       ELSE
!         MYLE= LENGTHK/2
!       ENDIF
!     ELSE
!       MYLE= NBLOKS
!     ENDIF

     L= MYLE*2

      print *,'I''m proc ',K,' inside _FACT my LENGTHK is ', LENGTHK ,&
             'my MYLE is ',MYLE,' and NFORPROC is', NFORPROC
!
! initialize JUMP, S
!
!  JUMP difference between index of block to reduce
!  and index of used factorization in same array MYMATR_A
!  S jump between processor wich are involved in communication
!
     JUMP= 1
     S= 1



!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!global reduction cycle!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! divided in two part :
! 1. GSTEP=1,..,NFORPROC , in this part each processor work
!      indipendently withou communication
! 2. GSTEP=NFORPROC+1,..,NFORPROC+LOGDUE(NPROCS),
!      in this LOGDUE(NPROCS) steps each processor send and receive
!      and after reduce its proper part

!Part 1
     DO GSTEP=1,NFORPROC

       !initialize MYHSTART
! index of next block to reduce
         MYHSTART= 2*EMPTY + 1

         ierr = MPE_Log_event( event6a, 0, "start compute-internal reduction")

         JUMP1= JUMP*2+1
!!internal reduction cycle!!
         DO I= 1,MYLE/2-1
!
! reshape [Ra Sb] as [Ra  and save in [ , ] columns by columns
!                      Sb]
           CALL REASSEMBLE( NRWBLK, 2, &
                            MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1) )
!
! update Z, MYINDEXP
!
           Z= MYHSTART+JUMP1
           MYINDEXP= MYINDEXP+1

      !     print *,'proc K', K, ' GSTEP', GSTEP,' I',I,'MYHSTART ',MYHSTART, &
      !             'JUMP ',JUMP, 'Z', Z

!
! reduce blocks  [Sa Ra   ] in [S'a R'a]
!                [   Sb Rb]
           CALL REDUCE_BLOCK( NRWBLK, MYMATR_A(:,:,MYHSTART), &
                      MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1), &
                      MYMATR_A(:,:,Z), MYPERM(:,MYINDEXP), &
                      MYFILL_IN(:,:,MYINDEXP), MYINFO )

           IF ( MYINFO.NE.0 ) THEN
             MYINFO= MYINDEXP
             RETURN
           ENDIF
!
! update MYHSTART
!
           MYHSTART= Z+1
         ENDDO

!separating last step from DO cycle is necessary
!only for processor NPROCS-1 when it has an odd number of blocks [Si Ri] to reduce

!!last step of internal reduction cycle
!
! reshape
!
         CALL REASSEMBLE( NRWBLK, 2, MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1) )
!
! update Z, MYINDEXP
!
           Z= MYHSTART+JUMP1
           MYINDEXP= MYINDEXP+1

!           IF ( Z.GT.L ) THEN
!             Z= L
!           ENDIF

          ! print *,'proc K', K, ' GSTEP', GSTEP,' I',I,'MYHSTART ',MYHSTART, &
          !         'JUMP ',JUMP, 'Z', Z

           CALL REDUCE_BLOCK( NRWBLK, MYMATR_A(:,:,MYHSTART), &
                    MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1), &
                    MYMATR_A(:,:,Z), MYPERM(:,MYINDEXP), &
                    MYFILL_IN(:,:,MYINDEXP), MYINFO )

           IF ( MYINFO.NE.0 ) THEN
              MYINFO= MYINDEXP
              RETURN
           ENDIF
!!end internal reduction cycle!!
         ierr = MPE_Log_event( event6b, 0, "end compute-internal reduction")
!
! update JUMP and MYLE
!
           JUMP= JUMP1
           MYLE= MYLE-MYLE/2


      ENDDO


!here L is even (2**NFORPROC)*2
      MYHSTART=2*EMPTY+1
      MYHEND= 2*EMPTY+L
      S=2
     ! LSTART=K
      PROD=MOD(K,S)
      JUMP=1


!Part 2
     DO GSTEP= 1,LOGNPROCS

        SENDARRAY(:,:,1)=MYMATR_A(:,:,MYHSTART)
        SENDARRAY(:,:,2)=MYMATR_A(:,:,MYHEND)

        IF (PROD==0) THEN
          !communication to K+S/2
          !tagsend is 1, tagrecv is 2
          OUT(GSTEP)=0
          print *,'K ', K,' before SENDRECV to proc',K+ S/2,' with GSTEP', &
          GSTEP, 'PROD', PROD, 'S', S

          ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MATR")
          CALL MPI_SENDRECV(SENDARRAY(1,1,1),2*NRWBLK*NRWBLK,MPI_DOUBLE_PRECISION, &
                            K+ S/2, GSTEP, MYMATR_A(1,1,MYHEND+1), 2*NRWBLK*NRWBLK, &
                            MPI_DOUBLE_PRECISION,K+ S/2,GSTEP,MPI_COMM_WORLD,STATUS, IERR )
          ierr = MPE_Log_event( event1b, 0, "end SENDRECV of MATR")


          MYHEND= MYHEND+2
          JUMP= L-1
        ELSE
          !communication to K-S/2
          OUT(GSTEP)=1
           print *,'K ', K,' before SENDRECV to proc',K- S/2,'with GSTEP', GSTEP,&
                   'PROD', PROD, 'S', S
          ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MATR")
          CALL MPI_SENDRECV(SENDARRAY(1,1,1),2*NRWBLK*NRWBLK,MPI_DOUBLE_PRECISION, &
                            K-S/2, GSTEP, MYMATR_A(1,1,MYHSTART-2), 2*NRWBLK*NRWBLK, &
                            MPI_DOUBLE_PRECISION,K-S/2,GSTEP,MPI_COMM_WORLD,STATUS, IERR )
          ierr = MPE_Log_event( event1b, 0, "start SENDRECV of MATR")


          MYHSTART=MYHSTART-2
          JUMP=1
        ENDIF

!update S
        S=S*2
!update PROD
        PROD=1
        DO I=0,2**(GSTEP)-1
          PROD= PROD*MOD(K-I,S)
        ENDDO
!        LSTART= LSTART - I
        ierr = MPE_Log_event( event6a, 0, "start compute-internal reduction")

!
! reshape [Ra Sb] as [Ra  and save in [ , ] columns by columns
!                      Sb]
           CALL REASSEMBLE( NRWBLK, 2, &
                            MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1) )

           MYINDEXP= MYINDEXP+1

!
! reduce blocks  [Sa Ra   ] in [S'a R'a]
!                [   Sb Rb]
           CALL REDUCE_BLOCK( NRWBLK, MYMATR_A(:,:,MYHSTART), &
                      MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1), &
                      MYMATR_A(:,:,MYHEND), MYPERM(:,MYINDEXP), &
                      MYFILL_IN(:,:,MYINDEXP), MYINFO )

           IF ( MYINFO.NE.0 ) THEN
             MYINFO= MYINDEXP
             RETURN
           ENDIF

           ierr = MPE_Log_event( event6b, 0, "end compute-internal reduction")

           L=L+2

     ENDDO

!!factorize  the last 2by 2
!fill the last block in MYMATR_A
     MYMATR_A(:,:,LENGTHK)= MYMATR_A(:,:,1)

!!
!! reshape the last blocks [R'''1 , Ba , Bb , S'''1]
!! as a [S''1 R''1 ]
!!      [ Ba   Bb  ]
!!
    CALL REASSEMBLE( NRWBLK, 4, MYMATR_A( :,:,LENGTHK-3:LENGTHK ) )

!!
!! factorize last matrix
!!
    MYINDEXP= MYINDEXP+1
    CALL DGETRF( NRWBLK*2, NRWBLK*2, MYMATR_A(1,1,LENGTHK-3), NRWBLK*2, &
                 MYPERM(1,(LENGTHK-3)/2), MYINFO )
    IF ( MYINFO.NE.0 ) THEN
      MYINFO= MYINDEXP
      STOP
    ENDIF


!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!end global reduction cycle!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!


      RETURN

        CONTAINS

          SUBROUTINE REDUCE_BLOCK( NRWBLK, S, RS, R, P, F, INFO )

            INTEGER,  INTENT (IN)     :: NRWBLK
            REAL(DP), INTENT(IN OUT)  :: RS( NRWBLK*2, NRWBLK ), &
                                         S( NRWBLK, NRWBLK ), R( NRWBLK, NRWBLK )
            REAL(DP), INTENT(OUT)     :: F( NRWBLK, NRWBLK )
            INTEGER,  INTENT(OUT)     :: P( NRWBLK*2 ), INFO

! Parameters
            REAL(DP), PARAMETER       :: ONE= 1.0D0, NONE= -1.0D0

! Local variables
            INTEGER                   :: I, J, T, P1( NRWBLK )
            REAL(DP)                  :: H( NRWBLK, NRWBLK ), L( NRWBLK, NRWBLK )

! Lapack routine : DGETRF ! Blas routines  : DTRSM, DGER


!
! factorize RS by means of the LU factorization
!
             CALL DGETRF( NRWBLK*2, NRWBLK, RS(1,1), NRWBLK*2, P1(1), INFO )
             IF ( INFO.NE.0 ) THEN
               RETURN
             ENDIF
!
! compute fill-in block F = C*L^(-1)
!
             CALL DTRSM( 'R', 'L', 'N', 'U', NRWBLK, NRWBLK, ONE, RS(1,1), &
                         NRWBLK*2, RS(NRWBLK+1,1), NRWBLK*2 )
             F= RS( NRWBLK+1:NRWBLK*2, : )
!
! determine the permutation vector
!
             DO I= 1, NRWBLK*2
               P( I )= I
             ENDDO
             DO I= 1, NRWBLK
               J= P1( I )
               IF ( J.NE.I ) THEN
                 T= P( I )
                 P( I )= P( J )
                 P( J )= T
               ENDIF
             ENDDO

!
! compute
! S' = S(1q)-F*S(1p), R' = R(2q)-F*R(2p)
! in three steps:
!
! first step,
! define H and L as blocks containing the nonnull rows of S(1q) and R(2q)
!
             H( :, : )= 0
             L( :, : )= 0
             DO I= 1, NRWBLK
               J= P( I+NRWBLK )
               IF ( J.LE.NRWBLK ) THEN
                 H( I, : )= S( J, : )
               ELSE
                 L( I, : )= R( J-NRWBLK, : )
               ENDIF
             ENDDO
!
! second step,
! compute  H = H-F*S(1p) and L = L-F*R(2p) by using
! knowledge of the nonnull rows of S(1p) and R(2p)
!
             DO I= 1, NRWBLK
               IF ( P(I).LE.NRWBLK ) THEN
! nonnull row of S(1p)
                 RS( NRWBLK+I, : )= S( P(I), : )
                 CALL DGER( NRWBLK, NRWBLK, NONE, F(1,I), 1, S(P(I),1), &
                            NRWBLK, H, NRWBLK )
               ELSE
! nonnull row of R(2p)
                 RS( NRWBLK+I, : )= R( P(I)-NRWBLK, : )
                 CALL DGER( NRWBLK, NRWBLK, NONE, F(1,I), 1, R(P(I)-NRWBLK,1), &
                            NRWBLK, L, NRWBLK )
               ENDIF
             ENDDO
!
! third step,
! set S' = H and R' = L
!
             S= H
             R= L

             RETURN
          END SUBROUTINE REDUCE_BLOCK

      END SUBROUTINE PARABABDCR_FACT4




 SUBROUTINE  PARABABDCR_SOLV4( K, LENGTHK, LENGTHVBK, NPROCS, LOGNPROCS, NFORPROC, &
                               NRWBLK, NBLOKS, MYMATR_A, &
                               MYPERM, MYFILL_IN, OUT, MYVECTB )
!
! Parameters
! ==========
!
! Input variables
!
!   K  integer,                  rank of processor in current communication domain
!                                K is in the set 0,...,NPROCS-1,
!
!   LENGTHK integer,             size of third dimension of MYMATR_A, LENGTHK>0
!
!   LENGTHVBK integer,           size of second dimension of MYVECTB, LENGTHVBK>0
!                                for K=0 is NBLOKS+1
!
!   NPROCS    integer,           number of processors, NPROCS>0
!
!   LOGNPROCS integer,           log2 of NPROCS
!
!   NFORPROC integer,            log2 of number of blocks [Si,Ri] in each proc
!                                in the last proc there is at most 2^NFORPROC + 1 blocks
!                                and at least 1 block, NFORPROC>0
!
!   NRWBLK integer,              the dimension of blocks S, R, Ba, Bb, NRWBLK>0
!
!   NBLOKS integer,              the number of blocks S(i) and R(i)
!                                in the babd matrix, NBLOKS>0
!
!   MYMATR_A double precision,   NRWBLK by NRWBLK by LENGTHK array,
!                                part of the babd matrix factorization
!
!   MYPERM   integer,            if K>0,
!                                 NRWBLK*2 by
!                                 (LENGTHK-2)/2= (LENGTHK-2 - LOGTOINT(K.EQ.0) )/2 array,
!                                 (since LOGTOINT(.TRUE.)=1, LOGTOINT(.FALSE.)=0)
!                                 the permutations associated to the
!                                 LU factorizations computed on proc K
!                                if K=0,
!                                 NRWBLK*2 by
!                                 (LENGTHK-3)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0))/2 array,
!                                 the permutations associated to the
!                                 LU factorizations computed on proc 0, even
!                                 associated to the factorization of the last
!                                 2 by 2 block system
!
!   MYFILL_IN double precision,  if K>0, NRWBLK by NRWBLK by
!                                 (LENGTHK-2)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
!                                 the fill-in matrix generated in the reduction
!                                 routine REDUCE_BLOCK called by proc K
!                                if K=0, NRWBLK by NRWBLK by
!                                 (LENGTHK-5)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
!                                 the fill-in matrix generated in the reduction
!                                 routine REDUCE_BLOCK called by proc 0
!
!  OUT  integer,      LOGNPROCS vector,
!         OUT(I) = 0 : processor K communicates to one processor with idle greater than K,
!                       at the I-th communication in the rhs reduction phase
!         OUT(I) = 1 : processor K communicates to one processor with idle lesser than K,
!                       at the I-th communication in the rhs reduction phase
!
!
!   MYVECTB double precision,     NRWBLK by LENGTHVBK array,
!                                 the right hand side of the babd linear system
!                                 stored in proc K
!
! Output variables
!   MYVECTB double precision,     NRWBLK by LENGTHVBK array,
!                                 the solution of the linear system
!
!


    INTEGER, INTENT(IN) :: K, LENGTHK, LENGTHVBK, NPROCS, LOGNPROCS, NFORPROC, &
                           NRWBLK, NBLOKS

!dimension of MYMATR_A, MYFILL_IN, MYPERM, MYVECTB changes through different processors


    REAL(DP), INTENT(IN) :: MYMATR_A( NRWBLK, NRWBLK, LENGTHK ), &
                            MYFILL_IN( NRWBLK, NRWBLK, (LENGTHK-5)/2  )


    INTEGER,  INTENT(IN) :: MYPERM( NRWBLK*2, (LENGTHK-3)/2), &
                            OUT(LOGNPROCS)

    REAL(DP), INTENT(IN OUT) :: MYVECTB( NRWBLK, LENGTHVBK )



! Local variables in each processor
    INTEGER :: MYINDEXP, MYLE, L, MYHSTART, JUMP2, S, &
               MYLEEQ2, MYHEND, GSTEP, INFO, LNEW, HNEW, &
               TAG, STATUS(MPI_STATUS_SIZE), IERR, &
               EMERGENCYTAG, EMERGENCYMSG(2), JUMP, P, DEST, &
               Z, I,  SOURCE, HSTART, LOLD, HOLD, J, H2, &
               H, EMPTY, event1a, event1b, event6a, event6b, &
               event3a, event3b

    REAL(DP) :: MSG(NRWBLK), TC(NRWBLK*2), XMSG(NRWBLK*2)


      event1a = MPE_Log_get_event_number()
      event1b = MPE_Log_get_event_number()
      event3a = MPE_Log_get_event_number()
      event3b = MPE_Log_get_event_number()
      event6a = MPE_Log_get_event_number()
      event6b = MPE_Log_get_event_number()


     ierr = MPE_Describe_state( event1a, event1b, &
                                "SendRecv MYVECTB in reduction phase", "green" )
     ierr = MPE_Describe_state( event6a, event6b, &
                                "Compute rhs reduction - internal cycle", "brown" )
     ierr = MPE_Describe_state( event3a, event3b, &
                                "Solve system", "blue" )




! Lapack routine : DGETRS
! Used subroutines: REDUCE_RHS, SOLVE_BLOCK

  EMPTY=SPACE(K+1,NPROCS)

!
! initialize MYINDEXP
!
     MYINDEXP= 0

!
! initialize MYLE, L
!
!  MYLE current number of rhs  of size NRWBLK to reduce
!   in proc 0, d(the rhs associated to boundary condition) is not considered
!   then MYLE is 2**NFORPROC, if K<NPROCS-1 and MYLE is LENGTHVBK-1, if K=NPROCS-1
!   (or NBLOKS, if NPROCS.EQ.1) at beginning
!  L current index of last nonnull rhs of size NRWBLK in MYVECTB
!   in proc 0,....,NPROCS-1, in proc 0 the rhs d is not considered
     MYLE= 2**NFORPROC
     L= MYLE

!
! initialize JUMP2
!
!  JUMP2 is twice JUMP that is the difference between indeces of rhs to reduce
!   in array MYVECTB
     JUMP2= 2


!
! initialize JUMP
!
!  JUMP difference between index of block to reduce
!  and index of used factorization in same array MYMATR_A
     JUMP= 1




!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!global reduction cycle!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! divided in two part :
! 1. GSTEP=1,..,NFORPROC , in this part each processor works
!      indipendently without communication
! 2. GSTEP=NFORPROC+1,..,NFORPROC+LOGDUE(NPROCS),
!      in this LOGDUE(NPROCS) steps each processor sends and
!      receives and after reduces its proper vector

!Part 1
     DO GSTEP=1,NFORPROC

      H= EMPTY + 1
    !  print *,'K ', K,' reduction rhs in GSTEP', GSTEP, &
    !          ' H',H,'JUMP', JUMP, 'NFORPROC',NFORPROC


      ierr = MPE_Log_event( event6a, 0, "start rhs compute-internal reduction")
!
!!!rhs internal reduction cycle!!!
!
         DO I= 1,MYLE/2
           MYINDEXP= MYINDEXP+1
!
! reduce ( Ta ) in T'a
!        ( Tb )
!

!           print *,'K ', K,' reduction rhs in GSTEP', GSTEP, &
!                   ' H',H,'JUMP', JUMP, ' I',I

           CALL REDUCE_RHS( NRWBLK, MYPERM(:,MYINDEXP), MYFILL_IN(:,:,MYINDEXP), &
                       MYVECTB(:,H), MYVECTB(:,H+JUMP) )
!
! update H
!
           H= H+JUMP2
         ENDDO
!
!!!end rhs internal reduction cycle!!!
!
         ierr = MPE_Log_event( event6b, 0, "end rhs compute-internal reduction")

         JUMP= JUMP2
         JUMP2= JUMP*2
         MYLE= MYLE- MYLE/2

      ENDDO

!here L is  (2**NFORPROC)
      MYHSTART= EMPTY+1
      MYHEND= EMPTY+L
      S=2
!here JUMP is the difference between the indeces of the
!two right hand side to reduce

! print *,'K',K,'After part1 in reduction VECTB '
!Part 2
     DO GSTEP= 1,LOGNPROCS


        IF (OUT(GSTEP)==0) THEN
          !communication to K+S/2
          !tagsend is 1, tagrecv is 2

          ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MYVECTB")
          CALL MPI_SENDRECV(MYVECTB(1,MYHSTART),NRWBLK,MPI_DOUBLE_PRECISION, &
                            K+ S/2, GSTEP, MYVECTB(1,MYHEND+1), NRWBLK, &
                            MPI_DOUBLE_PRECISION,K+S/2,GSTEP ,MPI_COMM_WORLD,STATUS,IERR )
          ierr = MPE_Log_event( event1b, 0, "start SENDRECV of MYVECTB")
          MYHEND= MYHEND+1
          JUMP= L
        ELSE
          !communication to K-S/2
          ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MYVECTB")
          CALL MPI_SENDRECV(MYVECTB(1,MYHSTART),NRWBLK,MPI_DOUBLE_PRECISION, &
                            K-S/2, GSTEP, MYVECTB(1,MYHSTART-1), NRWBLK, &
                            MPI_DOUBLE_PRECISION,K-S/2,GSTEP,MPI_COMM_WORLD,STATUS,IERR )
          ierr = MPE_Log_event( event1b, 0, "end SENDRECV of MYVECTB")
          MYHSTART=MYHSTART-1
          JUMP=1
        ENDIF

       ierr = MPE_Log_event( event6a, 0, "start rhs compute-internal reduction")

        MYINDEXP= MYINDEXP + 1
!
! reduce ( Ta ) in T'a
!        ( Tb )
!
           CALL REDUCE_RHS( NRWBLK, MYPERM(:,MYINDEXP), MYFILL_IN(:,:,MYINDEXP), &
                       MYVECTB(:,MYHSTART), MYVECTB(:,MYHSTART+JUMP) )


!
!!!end rhs internal reduction cycle!!!
!
         ierr = MPE_Log_event( event6b, 0, "end rhs compute-internal reduction")
           L=L+1
           S=S*2
     ENDDO


!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!end rhs global reduction cycle!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!



!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! 2 by 2 block linear system solution !!!!!!! to each processor
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! solve 2 by 2 block system
!!
!!           ( S^(1) R^(1) ) (x(0)     )= ( T^(1) )
!!           ( Ba      Bb  ) (x(NBLOKS))  (   d   )
!
  !     IF (K.EQ.0) THEN
        TC( 1:NRWBLK*2 )= (/ MYVECTB(1:NRWBLK,1), MYVECTB(1:NRWBLK,LENGTHVBK) /)
        CALL DGETRS( 'N', NRWBLK*2, 1, MYMATR_A(1,1,LENGTHK-3), &
                     NRWBLK*2, MYPERM(1,(LENGTHK-3)/2), TC, NRWBLK*2, INFO )
        MYVECTB( :, 1 )= TC( 1:NRWBLK )
        MYVECTB( :, LENGTHVBK )= TC( NRWBLK+1:NRWBLK*2 )

!        print *,'proc', K,'After solution of 2 by 2 MYVECTB', MYVECTB,' GSTEP=', GSTEP, 'MYLE',MYLE
  !   ENDIF



!
! initialize L
!  now L is the index of greater known x(l)
!  that will be used in solution phase
      L= LENGTHVBK
      H= 1
      LNEW= L
      HNEW= H
!
!!!!solution part
!!
!1) first part I= LOGNPROCS,-1,1
!
    ierr = MPE_Log_event( event3a, 0, "start solution system")
    DO GSTEP= LOGNPROCS,1,-1


     IF (OUT(GSTEP)==0) THEN
        J= L-1
        LNEW= J
      ELSE
        J= H+1
        HNEW= J
      ENDIF

      CALL SOLVE_BLOCK( NRWBLK, MYVECTB(:,H), &
                        MYVECTB(:,L), MYPERM(:,MYINDEXP), &
                        MYMATR_A(:,:,J*2-2:J*2-1), MYVECTB(:,J) )
      H=HNEW
      L=LNEW
      MYINDEXP=MYINDEXP-1
    ENDDO

!
!2) second part
!
! the start value for H and L here are HNEW and LNEW

    JUMP= (2**NFORPROC)/2
    H2= 1
    DO GSTEP= NFORPROC,1,-1

      DO I=1, H2
        J=L-JUMP

        CALL SOLVE_BLOCK( NRWBLK, MYVECTB(:,H), &
                        MYVECTB(:,L), MYPERM(:,MYINDEXP), &
                        MYMATR_A(:,:,J*2-2:J*2-1), MYVECTB(:,J) )
        MYINDEXP=MYINDEXP-1
        L=H
        H=H-JUMP*2
      ENDDO

      L=LNEW
      H= L- JUMP
      JUMP=JUMP/2
      H2= H2*2

    ENDDO


!!
!!!
!!!!!end solution part
!!!
!!
     ierr = MPE_Log_event( event3b, 0, "end solution system")
      RETURN

      CONTAINS

        SUBROUTINE REDUCE_RHS( NRWBLK, P, F, T1, T2 )
!
!  ==== BABDCR package ================
!  Giuseppe Romanazzi, Pierluigi Amodio
!  Dipartimento di Matematica
!  Universita' di Bari
!  June 20, 2004
!  ====================================
!
! Purpose
! =======
!
!  REDUCE_RHS  reduces a  NRWBLK*2 vector
!
!     ( T(1) )
!     ( T(2) )
!
!  in a NRWBLK vector T' by means of the
!  factorization in REDUCE_BLOCK.
!  This reduction uses fill-in matrix F
!  and permutation matrix P that comes out from
!  the factorization of the matrix
!
!    ( S(1) R(1)      )
!    (      S(2) R(2) )
!
!  to obtain T' = T(q)-F*T(p),
!  where
!
!        P*( T(1) ) = ( T(p) )
!          ( T(2) )   ( T(q) )
!
!  On exit, REDUCE_RHS gives the computed vector T' and T(p)
!
!
! Parameters
! ==========
!
! Input variables
!   NRWBLK integer,              the dimension of T1 and T2, NRWBLK>0
!
!   P      integer,              NRWBLK*2 vector,
!                                permutation vector that comes out
!                                from REDUCE_BLOCK
!
!   F     double precision,      NRWBLK by NRWBLK array,
!                                fill-in matrix that comes out
!                                from REDUCE_BLOCK
!
!   T1    double precision,      NRWBLK vector,
!                                first block row of input vector to reduce
!
!   T2    double precision,      NRWBLK vector,
!                                second block row of input vector to reduce
!
! Output variables
!   T1    double precision,      NRWBLK vector,
!                                the vector T' obtained in the reduction
!
!   T2    double precision,      NRWBLK  vector,
!                                the vector T(p) obtained by the permutation
!
          INTEGER,  INTENT(IN)     :: NRWBLK, P( NRWBLK*2 )
          REAL(DP), INTENT(IN)     :: F( NRWBLK, NRWBLK )
          REAL(DP), INTENT(IN OUT) :: T1( NRWBLK ), T2( NRWBLK )

! Parameters
          REAL(DP), PARAMETER       :: ONE= 1.0D0, NONE= -1.0D0

! Local variables
          INTEGER :: I, Q( NRWBLK*2 )
          REAL(DP):: TEMP( NRWBLK*2 )

! Blas routine: DGEMV

! Intrinsic procedure(not used at the moment)
!          INTRINSIC MATMUL

!
!define the inverse of P
!
           DO I= 1, NRWBLK*2
             Q( P( I ) )= I
           ENDDO
!
! permute vector ( T(1) )
!                ( T(2) )
!
           TEMP( Q( 1:NRWBLK ) )= T1( 1:NRWBLK )
           TEMP( Q( NRWBLK+1:NRWBLK*2 ) )= T2( 1:NRWBLK )
!
! determine T(p)
!
           T2= TEMP( 1:NRWBLK )
!
! compute T' = T(q)-F*T(p)

           T1= TEMP( NRWBLK+1:NRWBLK*2 )
           CALL DGEMV( 'N', NRWBLK, NRWBLK, NONE, F, NRWBLK, T2, 1, &
                       ONE, T1, 1 )

! the same result may be obtained by means of the intrinsic lapack procedure MATMUL
!
!            T1= TEMP( NRWBLK+1:NRWBLK*2 )-MATMUL( F, T2 )

           RETURN
        END SUBROUTINE REDUCE_RHS

        SUBROUTINE SOLVE_BLOCK( NRWBLK, X0, X2, P, LURS, T )
!
!  ==== BABDCR package ================
!  Giuseppe Romanazzi, Pierluigi Amodio
!  Dipartimento di Matematica
!  Universita' di Bari
!  June 20, 2004
!  ====================================
!
! Purpose
! =======
!
!  SOLVE_BLOCK  computes the solution x(1) of a linear system
!
!          L*U x(1) = T(p) - S(1p)*x(0) - R(2p)*x(2)
!
!  (of dimension NRWBLK) obtained from the subroutines REDUCE_BLOCK
!  and REDUCE_RHS applied to the system
!
!   ( S(1) R(1)      ) * ( x(0) ) = ( T(1) )
!   (      S(2) R(2) )   ( x(1) )   ( T(2) )
!                        ( x(2) )
!
!  Each block is NRWBLK by NRWBLK.
!
!  On exit, SOLVE_BLOCK gives the solution x(1)
!
!
! Parameters
! ==========
!
! Input variables
!   NRWBLK   integer,            the dimension of blocks of the original matrix, NRWBLK>0
!
!   XO     double precision,     NRWBLK vector, solution x(0) previously obtained
!
!   X2     double precision,     NRWBLK vector, solution x(2) previously obtained
!
!   P      integer,              NRWBLK*2 vector,
!                                permutation vector that comes out from REDUCE_BLOCK
!
!   LURS   double precision,     NRWBLK*2 by NRWBLK array,
!                                RS array that comes out from REDUCE_BLOCK
!
!   T      double precision,     NRWBLK vector, the rhs T(p)
!
!
! Output variables
!   T      double precision,     NRWBLK vector, the solution x(1)
!

          INTEGER,  INTENT (IN)     ::  NRWBLK, P( NRWBLK*2 )
          REAL(DP), INTENT (IN)     ::  X0( NRWBLK ), X2( NRWBLK ), &
                                          LURS( NRWBLK*2, NRWBLK )
          REAL(DP), INTENT (IN OUT) ::  T( NRWBLK )

! Local variables
          INTEGER :: I, INFO, P1( NRWBLK )

! Lapack routine: DGETRS

! Blas routine: DDOT

! Intrinsic procedure (not used in this implementation)
!          INTRINSIC DOT_PRODUCT

!
! compute T(p) - S(1p)*x(0) - R(2p)*x(2)
!
           DO I= 1, NRWBLK
             IF ( P(I).LE.NRWBLK ) THEN
!               T( I )= T( I ) - DOT_PRODUCT( LURS( NRWBLK+I,1:NRWBLK ),X0 )
                T( I )= T( I ) - DDOT( NRWBLK, LURS( NRWBLK+I, 1:NRWBLK ), 1, X0, 1)
             ELSE
!               T( I )= T( I ) - DOT_PRODUCT( LURS( NRWBLK+I,1:NRWBLK ),X2 )
                T( I )= T( I ) - DDOT( NRWBLK, LURS( NRWBLK+I, 1:NRWBLK ), 1, X2, 1)
             ENDIF
           ENDDO
!
! initialize identity array to be used in DGETRS
!
           DO I= 1, NRWBLK
             P1( I )= I
           ENDDO
!
! solve system L*U x(1) = (T(p)-S(1p)*x(0)-R(2p)*x(2))
!
           CALL DGETRS( 'N', NRWBLK, 1, LURS(1,1), 2*NRWBLK, &
   &                    P1(1), T(1), NRWBLK, INFO )

           RETURN
        END SUBROUTINE SOLVE_BLOCK


   END SUBROUTINE PARABABDCR_SOLV4


!!!!!!!!!!!!!!!!!!!!!!
!!in this work reassemble is used even in main2.f90
          SUBROUTINE  REASSEMBLE( NRWBLK, NBLOCK, MATR )

            INTEGER,  INTENT(IN)     :: NRWBLK, NBLOCK
            REAL(DP), INTENT(IN OUT) :: MATR( NRWBLK, NRWBLK*NBLOCK )

! Local variables
            REAL(DP) :: TEMP( NRWBLK*2, NRWBLK*NBLOCK/2 )

! Intrinsic procedure
            INTRINSIC  RESHAPE


             IF ( NBLOCK.EQ.2 ) THEN
               TEMP( 1:NRWBLK, : )= MATR( :, 1:NRWBLK )
               TEMP( NRWBLK+1:NRWBLK*2, : )= MATR( :, NRWBLK+1:NRWBLK*2 )
               MATR= RESHAPE( TEMP, (/ NRWBLK, NRWBLK*2 /) )
             ELSE
               TEMP( 1:NRWBLK, 1:NRWBLK )= MATR( :, NRWBLK*3+1:NRWBLK*4 )
               TEMP( 1:NRWBLK, NRWBLK+1:NRWBLK*2 )= MATR( :, 1:NRWBLK )
               TEMP( NRWBLK+1:NRWBLK*2, : )= MATR( :, NRWBLK+1:NRWBLK*3 )
               MATR= RESHAPE( TEMP, (/ NRWBLK, NRWBLK*4 /) )
             ENDIF

             RETURN
          END SUBROUTINE REASSEMBLE

!
!  !if L is .TRUE.  then LOGTOINT is 1
!  !   L is .FALSE. then LOGTOINT is 0
!          FUNCTION LOGTOINT(L)
!           INTEGER::  LOGTOINT
!           LOGICAL::  L
!
!            IF (L) THEN
!              LOGTOINT=1
!            ELSE
!              LOGTOINT=0
!            ENDIF
!
!            RETURN
!         END FUNCTION LOGTOINT
!
!!function logdue:  computes the less grater integer of log_2(x)
FUNCTION LOGDUE(X)
   IMPLICIT NONE
   INTEGER             :: LOGDUE, Y
   INTEGER, INTENT(IN) :: X
    LOGDUE= 0
    Y=X
    DO WHILE ( Y.NE.1 )
     LOGDUE=LOGDUE+1
     IF (MOD(Y,2).NE.0) THEN
      !here Y/2 is  floor of real Y/2
      Y= Y/2 + 1
     ELSE
      Y=Y/2
     ENDIF
    ENDDO
    RETURN
END FUNCTION LOGDUE
!
!FUNCTION LENGTH( K, NFORPROC, NPROCS, NGSTEPS, NBLOKS )
!  IMPLICIT NONE
!  INTEGER :: LENGTH, L, M
!  INTEGER, INTENT(IN) :: K, NFORPROC, NPROCS, NGSTEPS, NBLOKS
!!  INTERFACE
!!    FUNCTION BASIC(K, NPROCS,NBLOKS,NFORPROC)
!!      INTEGER             :: BASIC
!!      INTEGER, INTENT(IN) :: K, NPROCS,NBLOKS,NFORPROC
!!    END FUNCTION BASIC
!!   END INTERFACE
!   IF (NPROCS.EQ.1) THEN
!     LENGTH=2*(NBLOKS+1) + 1
!   ELSE
!     L= 2**NFORPROC
!     IF ( K.EQ.0 ) THEN
!       LENGTH= 3 + 2*(L + NGSTEPS - NFORPROC)
!     ELSE
!      IF ( K.EQ.NPROCS-1 ) THEN
!        LENGTH= 2*(NBLOKS-L*(NPROCS-1))
!      ELSE
!       LENGTH=BASIC(K,NPROCS, NBLOKS, NFORPROC)
!       LENGTH= 2*(L + LENGTH)
!      ENDIF
!     ENDIF
!   ENDIF
!   RETURN
!END FUNCTION LENGTH
!
!
!
!!determines the number of  block [Si Ri] which are added to
!!initial blocks in MYMATR_A of proc with rank K in MPI_COMM_WORLD
!! k=1,...,NPROCS-1
!FUNCTION BASIC(K, NPROCS, NBLOKS, NFORPROC)
!  IMPLICIT NONE
!  INTEGER             :: BASIC, M, S
!  INTEGER, INTENT(IN) :: K, NPROCS, NBLOKS, NFORPROC
!  IF ( MOD(K,2).EQ.0 ) THEN
!    M= K/2
!    BASIC= 1
!    S=2
!    DO WHILE (   ( (MOD(M,2).EQ.0).AND.(NPROCS.GE.K+S+1) ).OR. &
!                 ( (MOD(M,2).EQ.0).AND.(NPROCS.LT.K+S+1).AND. &
!                 (NBLOKS.GT.(2**NFORPROC)*NPROCS) )   )
!      M= M/2
!      BASIC= BASIC+1
!      S=S*2
!    ENDDO
!  ELSE
!    BASIC= 0
!  ENDIF
!  RETURN
!END FUNCTION BASIC

!this function determines the number of blank blocks of size
!2 x NRWBLK x NRWBLK  in MYMATR_A before the fill blocks
!received from proc 0 of the initial matrix.
FUNCTION SPACE(P,NPROCS)
  USE PRECISION
   INTEGER, INTENT(IN) :: P, NPROCS
   INTEGER :: SPACE, L, K, H, H2, V(NPROCS)

     V(1)=0
     V(2)=1
     H=1
     H2=2
     DO L=2, LOGDUE(NPROCS)
       H=H*2
       H2=H2*2
       DO K= H+1, H2
         V(K)= L - V(H2-K+1)
       ENDDO
     ENDDO
     SPACE=V(P)
END FUNCTION SPACE


END MODULE PARABABDCR_PACKAGE4
