!Paramodg.f90  - The Effective Cyclic Reduction
! is the general version  of paramod4.f90
MODULE PRECISION
  INTEGER, PARAMETER :: DP=KIND(1.0D0)
END MODULE PRECISION

MODULE PARABABDCR_PACKAGEG
  USE PRECISION
  USE MPI

  CONTAINS


    SUBROUTINE PARABABDCR_FACTG ( PROC, LENGTHM, LENGTHP, LENGTHF, NPROCS, &
                                  LOGNPROCS, NFORPROC, LOGNFORPROC, &
                                  k, m, NBLOKS, MYMATR_A, MYPERM, &
                                  MYFILL_IN, MYINFO, OUT )

!
! Parameters
! ==========
!
! Input variables
!
!   PROC  integer,               rank of processor in current communication domain
!                                PROC is in the set 0,...,NPROCS-1
!
!   LENGTHM integer,             length of MYMATR_A, LENGTHM>0
!
!   LENGTHP  integer,            length of MYPERM,
!
!   LENGTHF integer,             length of MYFILL_IN,
!
!   NPROCS    integer,           number of processors, NPROCS>0
!
!   LOGNPROCS integer,           log2 of NPROCS, is the number of the
!                                communicationo one-one with other preocessors
!
!   NFORPROC integer,            number of blocks [A(i),B(i),C(i)] in each proc
!                                in the last proc there is at most 2^NFORPROC + 1 blocks
!                                and at least 1 block, NFORPROC>0
!
!   LOGNFORPROC integer,         number of step of reduction before the communication,
!                                it is logdue(NFORPROC)
!
!   k,m integer,                 A(i) is (k+m)x m, B(i) is (k+m)xk, C(i) is (k+m)xm
!
!   NBLOKS integer,              the number of blocks [A(i), B(i), C(i)]
!                                in the babd matrix, NBLOKS>0
!
!   MYMATR_A double precision,   LENGTHM vector,
!                                the blocks of babd matrix which are saved
!                                in current processor
!
!   MYPERM integer,              LENGTHP vector,
!
!   MYFILL_IN double precision,  LENGTHF vector,
!
!
! Output variables
!   MYMATR_A double precision,   LENGTHM vector,
!                                part of the babd matrix factorization
!
!   MYPERM   integer,             LENGTHP vector
!                                 if K>0,
!                                 NRWBLK*2 by
!                                 (LENGTHK-2)/2= (LENGTHK-2 - LOGTOINT(K.EQ.0) )/2 array,
!                                 (since LOGTOINT(.TRUE.)=1, LOGTOINT(.FALSE.)=0)
!                                 the permutations associated to the
!                                 LU factorizations computed on proc K
!                                if K=0,
!                                 NRWBLK*2 by
!                                 (LENGTHK-3)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0))/2 array,
!                                 the permutations associated to the
!                                 LU factorizations computed on proc 0, even
!                                 associated to the factorization of the last
!                                 2 by 2 block system
!
!   MYFILL_IN double precision,  if K>0, NRWBLK by NRWBLK by
!                                 (LENGTHK-2)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
!                                 the fill-in matrix generated in the reduction
!                                 routine REDUCE_BLOCK called by proc K
!                                if K=0, NRWBLK by NRWBLK by
!                                 (LENGTHK-5)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
!                                 the fill-in matrix generated in the reduction
!                                 routine REDUCE_BLOCK called by proc 0
!
!   MYINFO   integer,
!          = 0:  successful exit
!          > 0:  if MYINFO = I, the I-th computed factorization
!                (in proc K) is performed on a
!                singular matrix; therefore the babd matrix is singular
!
!
!  OUT  integer,      LOGNPROCS vector,
!         OUT(I) = 0 : processor PROC communicates to one processor with idle greater than PROC,
!                       at the I-th communication in the matrix reduction phase
!         OUT(I) = 1 : processor PROC communicates to one processor with idle lesser than PROC,
!                       at the I-th communication in the matrix reduction phase
!


    INTEGER, INTENT(IN) :: PROC, LENGTHM, LENGTHP, LENGTHF, NPROCS, LOGNPROCS, &
                           NFORPROC, LOGNFORPROC, k, m, NBLOKS

!dimension of MYMATR_A, MYFILL_IN, MYPERM changes through different processors
    REAL(DP), INTENT(IN OUT) :: MYMATR_A( LENGTHM )

    REAL(DP), INTENT(OUT)    :: MYFILL_IN( m,LENGTHF )

    INTEGER,  INTENT(OUT)    :: MYPERM( LENGTHP ), &
                                MYINFO, OUT( LOGNPROCS )

! Local variables in each processor
    INTEGER ::  HSTART, HEND, IP, INF, JP, JF, JUMP, GSTEP, Z, PROD, I, S,&
                JUMP1, JUMP2, L, ierr, EMPTY, INFO, MYLE, m2, &
                STATUS(MPI_STATUS_SIZE) !, &
!                event1a,event1b, &
!                event6a,event6b
    REAL(DP) :: SENDARRAY(2*(m**2))

! Lapack routine : DGETRF
! Used subroutines: REASSEMBLE, REDUCE_BLOCK


!      event1a = MPE_Log_get_event_number()
!      event1b = MPE_Log_get_event_number()
!      event6a = MPE_Log_get_event_number()
!      event6b = MPE_Log_get_event_number()


!      ierr = MPE_Describe_state( event1a, event1b, &
!                                 "SendRecv MYMATR_A in reduction phase", "red" )
!      ierr = MPE_Describe_state( event6a, event6b, &
!                                 "Compute reduction - internal cycle", "yellow" )



      EMPTY=SPACE(PROC+1,NPROCS)

      m2=m**2
!
! initialize IP, INF, JP, JF
!
      IP=  1
      INF= 1
      JP=  (k+m)*2
      JF=  m+k*2

!
! initialize MYLE, L
!
!  MYLE current number of blocks [Ai Bi Ci] to reduce or of [A'i C'i]
!     MYLE= NFORPROC

!  L current index of last nonnull NRWBLK x NRWBLK block in MYMATR_A,
!   in proc 0 the blocks Ba,Bb are not considered and even initial
!   zero blocks
!     IF ( NPROCS.GT.1 ) THEN
!       IF ( K.LT.NPROCS-1 ) THEN
!       ELSE
!         MYLE= LENGTHK/2
!       ENDIF
!     ELSE
!       MYLE= NBLOKS
!     ENDIF
!     L= MYLE*2


!
! initialize JUMP, S
!
!  JUMP in the same reduction  step is the difference of two consecutive
!  couple block rows,
!  S jump between processor wich are involved in communication
!
     JUMP= (k+m)*(k+m*2)*2
     S= 1



!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!global reduction cycle!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!

!
! it is divided in two part :
! 1. GSTEP=1,..,LOGNFORPROC, in this part each processor work
!      indipendently without communication
! 2. GSTEP=NFORPROC+1,..,NFORPROC+LOGDUE(NPROCS),
!      in this LOGDUE(NPROCS) steps each processor send and receive
!      and after reduce its proper part

!Part 1
! In the fist step I reduce each couple
!    [Ai Bi Ci
!           Ai+1 Bi+1 Ci+1]
!
!    in [A'i C'i+1]
!

!
! initialize HSTART and HEND
!
! HSTART index of the first element of the block Ai where
!  Ai Bi Ci
!        Ai+1 Bi+1 Ci+1
! is to reduce
! or  of A'i  where A'i C'i
!                       A'i+1 C'i+1
! is to reduce
     HSTART= 2*(m2)*EMPTY+1


! HEND index of the first element of the last block to reduce
! that is C'i+1 in the last case examinited before,
     HEND= 2*(m2)*EMPTY+NFORPROC*(k+m)*(k+m*2)-(m2)+1

!I have NFORPROC couple
     DO I=1, NFORPROC/2

!
! reduce blocks  [Ai Bi Ci            ] in [A'i C'i+1]
!                [      Ai+1 Bi+1 Ci+1]
!

       CALL REDUCE_BLOCKG2( k, m, MYMATR_A(HSTART), &
                            MYPERM(IP), MYFILL_IN(1,INF), INFO )


       IF ( INFO.NE.0 ) THEN
         MYINFO= IP
         RETURN
       ENDIF
!
! update IP, INF
!
       IP= IP + JP
       INF= INF + JF
!
! update HSTART
!
       HSTART= HSTART+JUMP

     ENDDO

     JP= m*2
     JF= m
     MYLE= NFORPROC - NFORPROC/2

     DO GSTEP= 2,LOGNFORPROC

!initialize HSTART
! index of the first element of the next block couple to reduce
        HSTART= 2*(m2)*EMPTY+1

!        ierr = MPE_Log_event( event6a, 0, "start compute-internal reduction")

!!JUMP1
! difference between index of the first element of A'i
! and the first element of C'i+1

        JUMP1= JUMP*2-m2

!!JUMP2
! difference between index of block to reduce
! and index of used factorization in same array MYMATR_A
        JUMP2= JUMP-m2

!!internal reduction cycle!!
        DO I= 1,MYLE/2-1
!
! reshape [Ra Sb] as [Ra  and save in [ , ] columns by columns
!                      Sb]
           CALL REASSEMBLE( m, 2, &
                            MYMATR_A(HSTART+JUMP2) )
!
! update Z, MYINDEXP
!
           Z= HSTART+JUMP1

!
! reduce blocks  [Sa Ra   ] in [S'a R'a]
!                [   Sb Rb]
           CALL REDUCE_BLOCK( m, MYMATR_A(HSTART), &
                      MYMATR_A(HSTART+JUMP2), &
                      MYMATR_A(Z), MYPERM(IP), &
                      MYFILL_IN(1,INF), MYINFO )


           IF ( MYINFO.NE.0 ) THEN
             MYINFO= IP
             RETURN
           ENDIF

           IP= IP+JP
           INF= INF+JF
!
! update HSTART
!
           HSTART= Z+m2
         ENDDO

!separating last step from DO cycle is necessary
!only for processor NPROCS-1 when it has an odd number of blocks [Si Ri] to reduce

!!last step of internal reduction cycle
!
! reshape
!
         CALL REASSEMBLE( m, 2, &
                          MYMATR_A(HSTART+JUMP2) )
!
! update Z
!
           Z= HSTART+JUMP1

!           IF ( Z.GT.HEND ) THEN
!             Z= HEND
!           ENDIF


           CALL REDUCE_BLOCK( m, MYMATR_A(HSTART), &
                    MYMATR_A(HSTART+JUMP2), &
                    MYMATR_A(Z), MYPERM(IP), &
                    MYFILL_IN(1,INF), MYINFO )

           IF ( MYINFO.NE.0 ) THEN
              MYINFO= IP
              RETURN
           ENDIF

           IP= IP+JP
           INF= INF+JF

!!end internal reduction cycle!!
 !          ierr = MPE_Log_event( event6b, 0, "end compute-internal reduction")
!
! update JUMP and MYLE
!
           JUMP= JUMP*2
           MYLE= MYLE-MYLE/2


      ENDDO


!initialize HSTART, PROD, L
      HSTART=2*(m2)*EMPTY+1

      S=2
      PROD=MOD(PROC,S)

!L is the current length of the array without Ba Bb 0
      L=JUMP

!Part 2
      DO GSTEP= 1,LOGNPROCS

        SENDARRAY= (/ MYMATR_A(HSTART:HSTART+(m2)-1), MYMATR_A(HEND:HEND+m2-1) /)


        IF (PROD==0) THEN
          !communication to PROC+S/2

          OUT(GSTEP)=0

!          ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MATR")
          CALL MPI_SENDRECV(SENDARRAY(1),2*(m2),MPI_DOUBLE_PRECISION, &
                            PROC+ S/2, GSTEP, MYMATR_A(HEND+m2), 2*(m2), &
                            MPI_DOUBLE_PRECISION, PROC+ S/2,GSTEP,MPI_COMM_WORLD,STATUS, IERR )
!          ierr = MPE_Log_event( event1b, 0, "end SENDRECV of MATR")

!update HEND, JUMP2
!HEND is the index of last current block to reduce of type C'i+1
          HEND= HEND+2*(m2)

!!  JUMP2 difference between index of block to reduce
!  and index of used factorization in same array MYMATR_A
!I can move this before the update of HEND and use JUMP2=HEND
          JUMP2= L-m2

        ELSE
          !communication to PROC-S/2
          OUT(GSTEP)=1
          HSTART= HSTART-2*(m2)
!          ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MATR")
          CALL MPI_SENDRECV(SENDARRAY(1),2*(m2),MPI_DOUBLE_PRECISION, &
                            PROC-S/2, GSTEP, MYMATR_A(HSTART), 2*(m2), &
                            MPI_DOUBLE_PRECISION,PROC-S/2,GSTEP,MPI_COMM_WORLD,STATUS, IERR )
!          ierr = MPE_Log_event( event1b, 0, "start SENDRECV of MATR")

          JUMP2= m2
        ENDIF

!update S
        S=S*2

!update PROD
        PROD=1
        DO I=0,2**(GSTEP)-1
          PROD= PROD*MOD(PROC-I,S)
        ENDDO

!        ierr = MPE_Log_event( event6a, 0, "start compute-internal reduction")


!
! reshape [Ra Sb] as [Ra  and save in [ , ] columns by columns
!                      Sb]
        CALL REASSEMBLE( m, 2, &
                         MYMATR_A(HSTART+JUMP2) )



!
! reduce blocks  [Sa Ra   ] in [S'a R'a]
!                [   Sb Rb]
        CALL REDUCE_BLOCK( m, MYMATR_A(HSTART), &
                    MYMATR_A(HSTART+JUMP2), &
                    MYMATR_A(HEND), MYPERM(IP), &
                    MYFILL_IN(1,INF), MYINFO )

        IF ( MYINFO.NE.0 ) THEN
           MYINFO= IP
           RETURN
        ENDIF

        IP= IP+JP
        INF= INF+JF

!        ierr = MPE_Log_event( event6b, 0, "end compute-internal reduction")

        L= L+2*(m2)

     ENDDO

!!factorize  the last 2 by 2
!fill the last block in MYMATR_A
     MYMATR_A(LENGTHM-m**2+1:LENGTHM)= MYMATR_A(1:m**2)

!!
!! reshape the last blocks [R'''1 , Ba , Bb , S'''1]
!! as a [S''1 R''1 ]
!!      [ Ba   Bb  ]
!!

    CALL REASSEMBLE( m, 4, MYMATR_A(L-m2+1) )
!!
!! factorize last matrix
!!

    CALL DGETRF( m*2, m*2, MYMATR_A(L-m2+1), m*2, &
                 MYPERM(IP), MYINFO )

    IF ( MYINFO.NE.0 ) THEN
      MYINFO= IP
      STOP
    ENDIF


!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!end global reduction cycle!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!


      RETURN

        CONTAINS


         SUBROUTINE REDUCE_BLOCKG2( k, m, ARRAY1, P, F, INFO )
       USE PRECISION
!
!  ==== BABDCR_G  package ================
!  == subroutine REDUCE_BLOCKG2 =========
!
!  ====================================
!
! Purpose
! =======
!
!  REDUCE_BLOCKG2 reduces a 2(k+m) by 2(k+2m)  matrix
!         m    k    m    k    m
! k+m  ( A(1) B(1) C(1)           )
! k+m  (           A(2) B(2) C(2) )
!
!  in a m by 2m matrix
!        m   m
!    m ( A', C')
!
!  by means of the following operations:
!
!  if P * ( B(1) C(1)     ) = ( L ) * U,
!         (      A(2) C(2))   ( T )
!
!   where L and U are (k+m+k) by (k+m+k)
!
!     ( L1       )     ( U1 U2 0   )
!   L=( L2 L3    )   U=(    U3 U4  )
!     ( L4 L5 L6 )     (       U5  )
!   lower and upper triangular matrix, respectively
!
!   then
!                                          m     k+m+k   m
!    P * ( A(1) B(1) C(1)           ) = (A(1p)   L*U   C(2p)) k+m+k
!        (           A(2) B(2) C(2) )   (A(1q)   T*U   C(2q)) k+m+k
!
!
!  and
!
!    A' = A(1q) - F*A(1p),   C' = C(2q) - F*C(2p),
!
!  where F = T*L^(-1), .
!
!
!  On exit, REDUCE_BLOCKG gives the computed blocks A' and C',
!  the L\U factorization, the nonnull rows of A(1p) and C(2p)
!
!
!
! Parameters
! ==========
!
! Input variables
!   k         integer,            #columns of block B, #no-overlap columns, k>0
!
!   m         integer,            #columns ok blocks A and C,
!                                 #overlap columns of two successive row block, m>0
!
!
!   ARRAY1    double precision,   k+m by 2(k+m*2) array,
!                                 the array ( A(1) B(1) C(1) A(2) B(2) C(2) )
!
!
! Output variables
!   ARRAY1     double precision,  k+m by 2(k+m*2) array,
!                                in the first m^2 elements,
!                                there is A', after the
!                                factorization with tempL2,  the non null rows of A(1p),
!                                and C(2p), and in the last m^2
!                                element there is C'.
!
!   P      integer,              (k+m)*2 vector,
!                                permutation vector associated to the
!                                factorization , 1<=P(I)<=(k+m)*2
!
!
!   F      double precision,     m by (k+m+k) array,
!                                fill-in block obtained in the reduction
!
!
!   INFO   integer,
!          = 0:  successful exit
!          < 0:  if INFO = -I, the I-th argument of DGETRF calling had an illegal value
!          > 0:  if INFO = I, U(I,I) of factorization is exactly zero. The factorization
!                has been completed, but the factor U is exactly
!                singular, and division by zero will occur if it is used
!                to solve a system of equations.
!

            INTEGER,  INTENT (IN)     :: k, m
            REAL(DP), INTENT(IN OUT)  :: ARRAY1( k+m, (k+m*2)*2 )
            REAL(DP), INTENT(OUT)     :: F( m, k*2+m )
            INTEGER,  INTENT(OUT)     :: P( (k+m)*2 ), INFO

! Parameters
            REAL(DP), PARAMETER       :: ONE= 1.0D0, NONE= -1.0D0

! Local variables
            INTEGER                   :: I, J, T, P2( (k+m)*2 ), module, frac, &
                                         N, IR, IC, Q2(m*2+k), TEMP(k+m*2)
            REAL(DP)                  :: Mi2( k+m*2, k+m ), H( m, m ), &
                                         L( m, m ), TEMPL2(m,k), ARRAY2(k+m+k,m)



! Lapack routine : DGETRF
! Blas routines  : DTRSM, DGER



!
              P2=0
!first k steps of the LU factorization of [Bi Ci]=ARRAY1(1:k+m,m+1:m+k+2m+k)
              call dgetf2mod(k+m,k+m,k,ARRAY1(1,m+1),k+m,P2(1),INFO)
!
!now ARRAY1(1:k+m,m+1:2k+3m) is [  L1\U1         U2    ]
!                               [\tilde{L2}  \tilde{Ci}]

!
!create an array [\tilde{Ci} 0  ]
!                [ Bi+1     Ci+1]
!
     Mi2(:,:)=0
     Mi2(1:m,1:m)= ARRAY1(k+1:k+m,m+k+1:k+m*2)
     Mi2(m+1:k+m*2,1:k+m)= ARRAY1(1:k+m,k+m*2+1:k*2+m*3)
     call dgetf2(k+m*2,k+m,Mi2(1,1),k+m*2,P2(k+1),INFO)

!!!

!TEMP is used for obtaining TEMPL2 as in reduce_blockg.f90
    TEMP=0
    DO I=1,m
      TEMP(I)=I
    ENDDO



!
!! determine permutation vector for the second factorization
! that is Q2
    DO I= 1, k+m*2
      Q2( I )= I
    ENDDO
    DO I= 1, k+m
      J= P2( k+I )
      IF ( J.NE.I ) THEN
        T= Q2( I )
        Q2( I )= Q2( J )
        Q2( J )= T

!!adding operation for obtaining TEMPL2 as in reduce_blockg.f90
        T=TEMP(I)
        TEMP(I)=TEMP(J)
        TEMP(J)=T


      ENDIF
    ENDDO
!!
!


  DO I=1,k+m*2
    P2(k+I)=P2(k+I)+k
  ENDDO

!!
!! determine the permutation vector
!!
             DO I= 1, (k+m)*2
               P( I )= I
             ENDDO
             DO I= 1, k*2+m
               J= P2( I )
               IF ( J.NE.I ) THEN
                 T= P( I )
                 P( I )= P( J )
                 P( J )= T
               ENDIF
             ENDDO
!!!



!
! compute fill-in block F = T*L^(-1), where F=[F1 F2 F3], T=[T1 T2 T3]
! F1,T1 are m by k
! F2,T2 are m by m
! F3,T3 are m by k
!
! in three steps
!


!first step
! compute F3 from the relation F3L6=T3, F3 will go in place of T3
! in Mi2

           CALL DTRSM( 'R', 'L', 'N', 'U', m, k, ONE, Mi2(m+1,m+1), &
                       k+m*2, Mi2(k+m+1,m+1), k+m*2 )


!second step
! compute F2 using F2L3= T2 -F3 L5
!
! first I compute T2-F3L5 this will go in place of T2 in Mi2

           CALL DGEMM ( 'N', 'N', m, m, k, NONE, Mi2(m+k+1,m+1), &
                        k+m*2, Mi2(m+1,1), k+m*2, ONE, Mi2(k+m+1,1), k+m*2 )

           CALL DTRSM( 'R', 'L', 'N', 'U', m, m, ONE, Mi2(1,1), &
                       k+m*2, Mi2(k+m+1,1), k+m*2 )


!copy F2, F3 in F
           F(1:m,k+1:k*2+m)=Mi2(k+m+1:k+m*2,1:k+m)

!third step
! compute F1 using F1L1= T1 -F2L2 - F3L4
!
! first I compute T1 -F2L2 -F3L4 and save it in F(1:m,1:k)

  F(1:m,1:k)=0
  DO I=k+1,k+m
  !determine -F2L2 and move it in F(1:m,1:k)
    IF ( Q2(I-k).LE.m ) THEN
        CALL DGER( m, k, NONE, F(1,I), 1, ARRAY1(Q2(I-k)+k,m+1), &
                 k+m, F(1,1), m )
    ENDIF
  ENDDO


 !determine -F3L4 and add to  F(1:m,1:k)
  DO I=k+m+1,k*2+m
    IF ( Q2(I-k).LE.m ) THEN
       CALL DGER( m, k, NONE, F(1,I), 1, ARRAY1(Q2(I-k)+k,m+1), &
                  k+m, F(1,1), m )
    ENDIF
  ENDDO


!compute T1 + F(1:m,1:k) and move in F(1:m,1:k)
  DO I=k*2+m+1,(k+m)*2
    IF ( Q2(I-k).LE.m ) THEN
       F(I-(k*2+m),1:k)= F(I-(k*2+m),1:k)+ ARRAY1(Q2(I-k)+k,m+1:m+k)

    ENDIF
  ENDDO


! solve F1L1= F(1:m,1:k) and move in F(1:m,1:k)
           CALL DTRSM( 'R', 'L', 'N', 'U', m, k, ONE, ARRAY1(1,m+1), &
                       k+m, F(1,1), m )

  J=0
  TEMPL2=0
!
!!operation for obtaining TEMPL2 as in reduce_blockg only (L2 L4)
!
    DO I=1,k+m
     IF (TEMP(I).NE.0) THEN
        J=J+1
        TEMPL2(J,1:k)=ARRAY1(TEMP(I)+k,1+m:k+m)
     ENDIF
    ENDDO

    ARRAY1(k+1:k+m,m+1:k+m)=TEMPL2(1:m,1:k)
!!!




! compute
! A' = A(1q)-F*A(1p), C' = C(2q)-F*C(2p)
! in three steps:
!
! first step,
! define H and L as blocks containing the rows of A(1q) and C(2q)
!
             H( :, : )= 0
             L( :, : )= 0
             DO I= 1, m
               J= P( I+ k*2+m )
               IF ( J.LE.k+m ) THEN
                 H( I, : )= ARRAY1( J, 1:m )
               ELSE
                 L( I, : )= ARRAY1( J-(k+m), k*2+m*3+1:k*2+m*4 )
               ENDIF
             ENDDO
!
! second step,
! compute  H = H-F*A(1p) and L = L-F*C(2p) by using
! knowledge of the nonnull rows of A(1p) and C(2p)
!
!save non null rows of A(1p) and C(2p) of length (k+m+k)*m
! in the array ARRAY1 starting by position ARRAY1(1,2(k+m)+1) (and
! ending in position 2(k+m)+ (k+m+k)*m ) if I consider ARRAY1 as a vector
! columns by columns,
! I copy A(1p) C(2p) copy in ARRAY2 at beginning

             DO I= 1,k*2+m
               IF ( P(I).LE.k+m ) THEN

! save nonnull row of A(1p) in ARRAY1 starting from 1,2(k+m)+1
! copy  ARRAY1(P(I), 1:m)   (in ARRAY(J,N))
! at the moment I copy this in an temporanealy array ARRAY2(k+m,m*2))
                 ARRAY2(I,1:m)=ARRAY1(P(I),1:m)
                 CALL DGER( m, m, NONE, F(1,I), 1, ARRAY1(P(I),1), &
                            k+m, H, m )
               ELSE
! save nonnull row of C(2p) in ARRAY2

                 ARRAY2(I,1:m)=ARRAY1(P(I)-(k+m),k*2+m*3+1:k*2+m*4)

                 CALL DGER( m, m, NONE, F(1,I), 1, ARRAY1(P(I)-(k+m),k*2+m*3+1), &
                            k+m, L, m )
               ENDIF
             ENDDO

! third step,
! set A' = H and C' = L
! A' in the first m^2 elements of ARRAY1
!
            call dcopy (m**2,H(1,1),1,ARRAY1(1,1),1)
!
! and L=R' in the last m^2 elements of ARRAY1
!
        frac= (m**2)/(k+m)
        module= mod(m**2,k+m)
        IF (module.EQ.0) THEN
          IC=k*2+m*4-frac+1
          IR=1
       ELSE
          IC=k*2+m*4-frac
          IR=k+m-module+1
       ENDIF

       call dcopy (m**2,L(1,1),1,ARRAY1(IR,IC),1)

!
!copy ARRAY2(1:k+m+k,1:m)=(A1p C1p) in
! ARRAY1 starting by position (1,(k+m)*2+1), columns by columns
!that is as ARRAY2 was a vector
!
     call dcopy( m*(m+2*k), ARRAY2(1,1), 1, ARRAY1(1,(k+m)*2+1),1)



!copy L1\U1 and TEMPL2
       J=module+1
       N=frac+1

       call dcopy((k+m)*k,ARRAY1(1,m+1),1,ARRAY1(J,N),1)

       N=N+k
!
!copy (   U2  ) in ARRAY1
!     ( L3\U3 )
!     (   L5  )
!
!first I move U2 in ARRAY2(1:k,1:m)
     ARRAY2(1:k,1:m)=ARRAY1(1:k ,k+m+1:k+m*2)
!copy L3\U3 L5 in ARRAY2(k+1:k*2+m,1:m)
     ARRAY2(k+1:k*2+m,1:m)=Mi2(1:k+m,1:m)

!and then I move ARRAY2 in ARRAY1
     call dcopy((k+m+k)*m, ARRAY2(1,1), 1, ARRAY1(J,N),1)


!
!copy (  U4   )= ARRAY(k+1:k*2+m,1:k)
!     ( L6\U5 )
!
     J= k+m*2+1
     DO I=1,k
       call dcopy(k+m, Mi2(1,m+I), 1, ARRAY1(1,J),1)
       J= J+1
     ENDDO

     RETURN
END SUBROUTINE REDUCE_BLOCKG2


          SUBROUTINE REDUCE_BLOCK( NRWBLK, S, RS, R, P, F, INFO )

            INTEGER,  INTENT (IN)     :: NRWBLK
            REAL(DP), INTENT(IN OUT)  :: RS( NRWBLK*2, NRWBLK ), &
                                         S( NRWBLK, NRWBLK ), R( NRWBLK, NRWBLK )
            REAL(DP), INTENT(OUT)     :: F( NRWBLK, NRWBLK )
            INTEGER,  INTENT(OUT)     :: P( NRWBLK*2 ), INFO

! Parameters
            REAL(DP), PARAMETER       :: ONE= 1.0D0, NONE= -1.0D0

! Local variables
            INTEGER                   :: I, J, T, P1( NRWBLK )
            REAL(DP)                  :: H( NRWBLK, NRWBLK ), L( NRWBLK, NRWBLK )

! Lapack routine : DGETRF ! Blas routines  : DTRSM, DGER


!
! factorize RS by means of the LU factorization
!
             CALL DGETRF( NRWBLK*2, NRWBLK, RS(1,1), NRWBLK*2, P1(1), INFO )
             IF ( INFO.NE.0 ) THEN
               RETURN
             ENDIF
!
! compute fill-in block F = C*L^(-1)
!
             CALL DTRSM( 'R', 'L', 'N', 'U', NRWBLK, NRWBLK, ONE, RS(1,1), &
                         NRWBLK*2, RS(NRWBLK+1,1), NRWBLK*2 )
             F= RS( NRWBLK+1:NRWBLK*2, : )
!
! determine the permutation vector
!
             DO I= 1, NRWBLK*2
               P( I )= I
             ENDDO
             DO I= 1, NRWBLK
               J= P1( I )
               IF ( J.NE.I ) THEN
                 T= P( I )
                 P( I )= P( J )
                 P( J )= T
               ENDIF
             ENDDO

!
! compute
! S' = S(1q)-F*S(1p), R' = R(2q)-F*R(2p)
! in three steps:
!
! first step,
! define H and L as blocks containing the nonnull rows of S(1q) and R(2q)
!
             H( :, : )= 0
             L( :, : )= 0
             DO I= 1, NRWBLK
               J= P( I+NRWBLK )
               IF ( J.LE.NRWBLK ) THEN
                 H( I, : )= S( J, : )
               ELSE
                 L( I, : )= R( J-NRWBLK, : )
               ENDIF
             ENDDO
!
! second step,
! compute  H = H-F*S(1p) and L = L-F*R(2p) by using
! knowledge of the nonnull rows of S(1p) and R(2p)
!
             DO I= 1, NRWBLK
               IF ( P(I).LE.NRWBLK ) THEN
! nonnull row of S(1p)
                 RS( NRWBLK+I, : )= S( P(I), : )
                 CALL DGER( NRWBLK, NRWBLK, NONE, F(1,I), 1, S(P(I),1), &
                            NRWBLK, H, NRWBLK )
               ELSE
! nonnull row of R(2p)
                 RS( NRWBLK+I, : )= R( P(I)-NRWBLK, : )
                 CALL DGER( NRWBLK, NRWBLK, NONE, F(1,I), 1, R(P(I)-NRWBLK,1), &
                            NRWBLK, L, NRWBLK )
               ENDIF
             ENDDO
!
! third step,
! set S' = H and R' = L
!
             S= H
             R= L

             RETURN
          END SUBROUTINE REDUCE_BLOCK

      END SUBROUTINE PARABABDCR_FACTG




 SUBROUTINE  PARABABDCR_SOLVG  ( PROC, LENGTHM, LENGTHP, LENGTHF, LENGTHVB, &
                                  NPROCS, LOGNPROCS, NFORPROC, LOGNFORPROC, &
                                  k, m, NBLOKS, MYMATR_A, MYPERM, &
                                  MYFILL_IN, OUT, MYVECTB )
!
! Parameters
! ==========
!
! Input variables
!
!   PROC  integer,               rank of processor in current communication domain
!                                PROC is in the set 0,...,NPROCS-1
!
!   LENGTHM integer,             length of MYMATR_A, LENGTHM>0
!
!   LENGTHP  integer,            length of MYPERM,
!
!   LENGTHF integer,             length of MYFILL_IN,
!
!   NPROCS    integer,           number of processors, NPROCS>0
!
!   LOGNPROCS integer,           log2 of NPROCS, is the number of the
!                                communicationo one-one with other preocessors
!
!   NFORPROC integer,            number of blocks [A(i),B(i),C(i)] in each proc
!                                in the last proc there is at most 2^NFORPROC + 1 blocks
!                                and at least 1 block, NFORPROC>0
!
!   LOGNFORPROC integer,         number of step of reduction before the communication,
!                                it is logdue(NFORPROC)
!
!   k,m integer,                 A(i) is (k+m)x m, B(i) is (k+m)xk, C(i) is (k+m)xm
!
!   NBLOKS integer,              the number of blocks [A(i), B(i), C(i)]
!                                in the babd matrix, NBLOKS>0
!
!   MYMATR_A double precision,   LENGTHM vector,
!                                part of the factorization of the input babd matrix
!
!   MYPERM integer,              LENGTHP vector, permutation vector associated
!                                to the factorization in MYMATR_A
!
!   MYFILL_IN double precision,  LENGTHF vector, fill-in vector associated
!                                to the factorization in MYMATR_A
!
!
!  OUT  integer,      LOGNPROCS vector,
!         OUT(I) = 0 : processor K communicates to one processor with idle greater than K,
!                       at the I-th communication in the rhs reduction phase
!         OUT(I) = 1 : processor K communicates to one processor with idle lesser than K,
!                       at the I-th communication in the rhs reduction phase
!

!   MYVECTB double precision,    LENGTHVB right hand side associated to MYMATR_A
!                                 the right hand side of the babd linear system
!                                 stored in proc PROC
!
! Output variables
!
!   MYVECTB double precision,    LENGTHVB solution part of the final solution
!



    INTEGER, INTENT(IN) :: PROC, LENGTHM, LENGTHF, LENGTHP, LENGTHVB, &
                           NPROCS, LOGNPROCS, NFORPROC, &
                           LOGNFORPROC, k, m, NBLOKS

!dimension of MYMATR_A, MYFILL_IN, MYPERM, MYVECTB changes through different processors


    REAL(DP), INTENT(IN) :: MYMATR_A( LENGTHM ), &
                            MYFILL_IN( m,LENGTHF )


    INTEGER,  INTENT(IN) :: MYPERM( LENGTHP ), &
                            OUT(LOGNPROCS)

    REAL(DP), INTENT(IN OUT) :: MYVECTB( LENGTHVB )



! Local variables in each processor
    INTEGER :: EMPTY, MYLE, &
               IP, JP, INF, JF, &
               H, J, L, HSTART, HEND, &
               JUMP, JUMP1, JUMP2, &
               S, INFO, GSTEP, HNEW, LNEW,  &
               STATUS(MPI_STATUS_SIZE), IERR, &
               I, &
               LLUV, HLUV, JLUV, LLUVOLD !, &
!               event1a, event1b, event6a, event6b, &
!               event3a, event3b

    REAL(DP) :: TC(m*2)


!      event1a = MPE_Log_get_event_number()
!      event1b = MPE_Log_get_event_number()
!      event3a = MPE_Log_get_event_number()
!      event3b = MPE_Log_get_event_number()
!      event6a = MPE_Log_get_event_number()
!      event6b = MPE_Log_get_event_number()


!     IERR = MPE_Describe_state( event1a, event1b, &
!                                "SendRecv MYVECTB in reduction phase", "green" )
!     IERR = MPE_Describe_state( event6a, event6b, &
!                                "Compute rhs reduction - internal cycle", "brown" )
!     IERR = MPE_Describe_state( event3a, event3b, &
!                                "Solve system", "blue" )




! Lapack routine : DGETRS
! Used subroutines: REDUCE_RHS, SOLVE_BLOCK

     EMPTY= SPACE(PROC+1,NPROCS)

!
! initialize IP, INF, JP, JF
!
      IP=  1
      INF= 1
      JP=  (k+m)*2
      JF=  m+k*2


!
! initialize JUMP, S
!
!  JUMP in the same reduction  step is the difference
!  of two consecutive couple of right hand sides,
!  in the first NFORPROC/2 reductions
!  S jump between processor wich are involved in communication
!
     JUMP= (k+m)*2
     S= 1



!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!rhs global reduction cycle!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!

!
! it is divided in two part :
! 1. GSTEP=1,..,LOGNFORPROC, in this part each processor work
!      indipendently without communication
! 2. GSTEP=NFORPROC+1,..,NFORPROC+LOGDUE(NPROCS),
!      in this LOGDUE(NPROCS) steps each processor send and receive
!      and after reduce its proper part

!Part 1
! In the fist step I reduce each couple
!         [Ti    k+m
!          Ti+1] k+m
!
!    in  (T'i) m
!

!
! initialize HSTART and HEND
!
! HSTART index of the first element of the block Ti where
!        Ti     k+m
!        Ti+1   k+m
! is to reduce
! or  of T'i  where  T'i     m
!                    T'i+1   m
! is to reduce
     HSTART= m*EMPTY+1

! HEND index of the first element of the last rhs to reduce
! that is of T'i+1 in the last cases examinited before,
     HEND= HSTART+NFORPROC*(k+m)-m

!I have NFORPROC couple
     DO I=1, NFORPROC/2
!
! reduce ( Ta ) in T'a
!        ( Tb )
!


       CALL REDUCE_RHSG( k, m, MYPERM(IP), MYFILL_IN(1,INF), MYVECTB(HSTART) )


!
! update IP, INF
!
       IP= IP + JP
       INF= INF + JF
!
! update HSTART
!
       HSTART= HSTART+JUMP


     ENDDO


     JP= m*2
     JF= m
     MYLE= NFORPROC - NFORPROC/2


     DO GSTEP= 2,LOGNFORPROC

!initialize HSTART
! index of the first element of the next block couple to reduce
       HSTART= m*EMPTY+1



!!  JUMP2 difference between two couple of rhs to reduce
       JUMP2=JUMP*2

!       IERR = MPE_Log_event( event6a, 0, "start rhs compute-internal reduction")

!
!!!rhs internal reduction cycle!!!
!
       DO I= 1,MYLE/2

!
! reduce ( Ta ) in T'a
!        ( Tb )
!

           CALL REDUCE_RHS( m, MYPERM(IP), MYFILL_IN(1,INF), &
                       MYVECTB(HSTART), MYVECTB(HSTART+JUMP) )

!
!update IP,INF
!
           IP= IP+JP
           INF= INF+JF

!
! update HSTART
!
           HSTART= HSTART+JUMP2
         ENDDO
!
!!!end rhs internal reduction cycle!!!
!
!         IERR = MPE_Log_event( event6b, 0, "end rhs compute-internal reduction")

         JUMP= JUMP2

         MYLE= MYLE- MYLE/2

      ENDDO


!initialize HSTART, PROD, L
      HSTART=m*EMPTY+1

      S=S*2

!L is the current length of the array, if we esclude Ba Bb 0
      L=JUMP

!in the next JUMP is the difference in the indece between
!two right hand side to reduce

!Part 2
     DO GSTEP= 1,LOGNPROCS

       IF ( OUT(GSTEP)==0 ) THEN
          !communication to PROC+S/2

  !       IERR = MPE_Log_event( event1a, 0, "start SENDRECV of MYVECTB")

         HEND= HEND+m
         CALL MPI_SENDRECV(MYVECTB(HSTART),m,MPI_DOUBLE_PRECISION, &
                           PROC+ S/2, GSTEP, MYVECTB(HEND), m, &
                           MPI_DOUBLE_PRECISION,PROC+S/2,GSTEP ,MPI_COMM_WORLD,STATUS,IERR )
 !        IERR = MPE_Log_event( event1b, 0, "start SENDRECV of MYVECTB")
         JUMP= L
       ELSE
          !communication to PROC-S/2
!         IERR = MPE_Log_event( event1a, 0, "start SENDRECV of MYVECTB")

         CALL MPI_SENDRECV(MYVECTB(HSTART),m,MPI_DOUBLE_PRECISION, &
                           PROC-S/2, GSTEP, MYVECTB(HSTART-m), m, &
                           MPI_DOUBLE_PRECISION,PROC-S/2,GSTEP,MPI_COMM_WORLD,STATUS,IERR )
 !        IERR = MPE_Log_event( event1b, 0, "end SENDRECV of MYVECTB")
         HSTART= HSTART-m
         JUMP= m
       ENDIF

 !      IERR = MPE_Log_event( event6a, 0, "start rhs compute-internal reduction")


!
! reduce ( Ta ) in T'a
!        ( Tb )
!
       CALL REDUCE_RHS( m, MYPERM(IP), MYFILL_IN(1,INF), &
                        MYVECTB(HSTART), MYVECTB(HSTART+JUMP) )

! IERR = MPE_Log_event( event6b, 0, "end rhs compute-internal reduction")
           IP= IP+JP
           INF= INF+JF
!
!!!end rhs internal reduction cycle!!!
!

           L= L+m
           S=S*2
     ENDDO


!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!end rhs global reduction cycle!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!


!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! 2 by 2 block linear system solution !!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! solve 2 by 2 block system
!!
!!           ( S^(1) R^(1) ) (x(0)     )= ( T^(1) )
!!           ( Ba      Bb  ) (x(NBLOKS))  (   d   )
!

        TC( 1:m*2 )= (/MYVECTB(1:m), MYVECTB(LENGTHVB-m+1:LENGTHVB)/)
        CALL DGETRS( 'N', m*2, 1, MYMATR_A(LENGTHM-4*(m**2)+1), &
                     m*2, MYPERM(IP), TC, m*2, INFO )
        MYVECTB( 1:m )= TC( 1:m )
        MYVECTB( LENGTHVB-m+1:LENGTHVB )= TC( m+1:m*2 )



!
! initialize L
!  now L is the index of greater known x(l)
!  that will be used in solution phase
      L= LENGTHVB-m+1

!LLUV index of the greater LUV current  used

      LLUV= LENGTHM-(m**2)*4 + 1

!HLUV index of the smaller LUV current  used
      HLUV= -(m**2)+1

      H= 1
      LNEW= L
      HNEW= H
!
!!!!solution part
!!
!1) first part I= LOGNPROCS,-1,1
!

!    ierr = MPE_Log_event( event3a, 0, "start solution system")
    DO GSTEP= LOGNPROCS,1,-1

     IP= IP-JP

     IF (OUT(GSTEP)==0) THEN
        J= L-m
        JLUV= LLUV-(m**2)*2
        LLUV= JLUV
        LNEW= J
     ELSE
        J= H+m
        JLUV= HLUV+(m**2)*2
        HLUV= JLUV
        HNEW= J
     ENDIF

      CALL SOLVE_BLOCK( m, MYVECTB(H), &
                       MYVECTB(L), MYPERM(IP), &
                       MYMATR_A(JLUV), MYVECTB(J) )

      H= HNEW
      L= LNEW

    ENDDO

!
!2) second part
!
! the start value for H and L here are HNEW and LNEW
    LLUVOLD=LLUV
    JUMP= (NFORPROC)/2

    MYLE= 1
    DO GSTEP= LOGNFORPROC,2,-1

      DO I=1, MYLE

        J= L- (JUMP/2)*m - (JUMP/2)*(k*2+m)

        JLUV= LLUV- (JUMP/2)*(m**2)*2 -(JUMP/2)*( (k+m)*(k+m*2)*2-(m**2)*2 )

        IP= IP-JP


        CALL SOLVE_BLOCK( m, MYVECTB(H), &
                          MYVECTB(L), MYPERM(IP), &
                          MYMATR_A(JLUV), MYVECTB(J) )

        L= H
        H= H-JUMP*m - (k*2+m)*JUMP
        LLUV= JLUV - (JUMP/2)*( (k+m)*(k+m*2)*2 )

        IF (H.LT.0) THEN
          H= 1
        ENDIF
! it's better write the last iterate out from the cycle
! since this selection is true always  for I=H2
      ENDDO

      L= LNEW
      H= L-(JUMP/2)*m -(JUMP/2)*(k*2+m)
      LLUV= LLUVOLD
      JUMP= JUMP/2
      MYLE= MYLE*2

    ENDDO
!last NFORPROC/2 step of solution, here H2=NFORPROC/2

    JP= (k+m)*2
    DO I= 1,MYLE

        J= L-(k*2+m)
        JLUV= LLUV- ( (k+m)*(k+m*2)*2-(m**2)*2 )

        IP= IP-JP


        CALL SOLVE_BLOCKG( k, m, MYVECTB(H), &
               MYVECTB(L), MYPERM(IP), &
               MYMATR_A(JLUV), MYVECTB(J) )

        LLUV= JLUV - (m**2)*2
        L= H
        H= H-(k*2+m)-m
    ENDDO

!!
!!!
!!!!!end solution part
!!!
!!
!     ierr = MPE_Log_event( event3b, 0, "end solution system")
      RETURN

      CONTAINS

       SUBROUTINE REDUCE_RHSG( k, m, P, F, T )
!
!  ==== BABDCR_G package ================
!  == subroutine REDUCE_RHSG ===========
!  ====================================
!
! Purpose
! =======
!
!  REDUCE_RHSG  reduces a  (k+m)*2 vector
!
!    T= ( T1 )
!       ( T2 )
!
!  in a m vector T' by means of the
!  factorization in REDUCE_BLOCKG.
!  This reduction uses fill-in matrix F
!  and permutation matrix P that comes out from
!  the reduction of the matrix
!
!    ( A(1) B(1) C(1)           )
!    (           A(2) B(2) C(2) )
!
!  to obtain T' = Tq-F*Tp,
!  where
!
!        P*( T1 ) = ( Tp ) k*2+m
!          ( T2 )   ( Tq ) m
!
!  On exit, REDUCE_RHSG gives the computed vector T' and Tp.
!
!
! Parameters
! ==========
!
! Input variables
!   k, m integer,              the dimension of T1 and T2, k, m>0
!
!   P      integer,              (k+m)*2 vector,
!                                permutation vector that comes out
!                                from REDUCE_BLOCKG
!
!   F     double precision,      m by k*2+m array,
!                                fill-in matrix that comes out
!                                from REDUCE_BLOCKG
!
!   T    double precision,       (k+m)*2 vector,
!                                the input vector to reduce
!
! Output variables
!   T    double precision,       (k+m)*2 vector,
!                                the vector (T') obtained in the reduction
!                                           (Tp)
!
          INTEGER,  INTENT(IN)     :: k, m, P( (k+m)*2 )
          REAL(DP), INTENT(IN)     :: F( m, k*2+m )
          REAL(DP), INTENT(IN OUT) :: T( (k+m)*2 )

! Parameters
          REAL(DP), PARAMETER       :: ONE= 1.0D0, NONE= -1.0D0

! Local variables
          INTEGER :: I, Q( (k+m)*2 )
          REAL(DP):: TEMP( (k+m)*2 )

! Blas routine: DGEMV

! Intrinsic procedure(not used at the moment)
!          INTRINSIC MATMUL

!
!define the inverse of P
!
           DO I= 1, (k+m)*2
             Q( P( I ) )= I
           ENDDO
!
! permute vector ( T(1) )
!                ( T(2) )
!
        TEMP( Q( 1:(k+m)*2 ) )= T( 1:(k+m)*2 )
!
! determine T(p)
!
           T(m+1:(k+m)*2)= TEMP( 1:k*2+m )
!
! compute T' = Tq-F*Tp
!
           T(1:m)= TEMP( k*2+m+1:(k+m)*2 )
           CALL DGEMV( 'N', m, k*2+m, NONE, F, m, T(m+1), 1, &
                       ONE, T(1), 1 )


           RETURN
        END SUBROUTINE REDUCE_RHSG


        SUBROUTINE SOLVE_BLOCKG( k, m, X0, X2, P, LURS, T )
!
!  ==== BABDCR_G package ================
!  == subroutine SOLVE_BLOCKG ==========
!  ====================================
!
! Purpose
! =======
!
!  SOLVE_BLOCKG  computes the solution x1 of a linear system
!
!          L*U x1 = Tp - A(1p)*x0 - C(2p)*x2
!
!  (of dimension k*2+m) obtained from the subroutines REDUCE_BLOCKG
!  and REDUCE_RHSG applied to the system
!        m    k    m    k     m
! k+m  ( A(1) B(1) C(1)           ) * ( x0 ) = ( T1 )
! k+m  (           A(2) B(2) C(2) )   ( x1 )   ( T2 )
!                                     ( x2 )
!
!  x0,x2 are of length, m x1 is of length k*2+m.
!  The rhs Tp is obtained after a row permutation of T1 and T2.
!
!  On exit, SOLVE_BLOCKG gives the solution x1.
!
!
! Parameters
! ==========
!
! Input variables
!   k,m    integer,              the dimensions of blocks of the original matrix, NRWBLK>0
!
!   XO     double precision,     m vector, the solution x(0) previously obtained
!
!   X2     double precision,     m vector, the solution x(2) previously obtained
!
!   P      integer,              (k+m)*2 vector,
!                                permutation vector that comes out from REDUCE_BLOCK
!
!   LURS   double precision,      by NRWBLK array,
!                                RS array that comes out from REDUCE_BLOCK
!
!   T      double precision,     NRWBLK vector, the rhs Tp
!
!
! Output variables
!   T      double precision,     NRWBLK vector, the solution x1
!

          INTEGER,  INTENT (IN)     ::  k,m , P( (k+m)*2 )
          REAL(DP), INTENT (IN)     ::  X0( m ), X2( m ), &
                                        LURS( (k*2+m)*(k+m*2)+k*m )

          REAL(DP), INTENT (IN OUT) ::  T( k*2+m )

! Local variables
          INTEGER :: I, INFO, P1( k*2+m ), J
          REAL(DP) :: DDOT, RS(k*2+m,m), LU(k*2+m,k*2+m), TEMP(k+m,k)

! Lapack routine: DGETRS

! Blas routine: DDOT

! Intrinsic procedure (not used in this implementation)
!          INTRINSIC DOT_PRODUCT

!copy   ( A1p ) in RS
!       ( C2p )
           call dcopy ((k*2+m)*m, LURS((k+m)*(k*2+m)+k*m+1),1,RS(1,1),1 )

!
! compute T(p) - A(1p)*x(0) - C(2p)*x(2)
!
           DO I= 1, k*2+m
             IF ( P(I).LE.k+m ) THEN
!               T( I )= T( I ) - DOT_PRODUCT( RS( I,1:m ), X0 )
                T( I )= T( I ) - DDOT( m, RS( I, 1:m ), 1, X0, 1)
             ELSE
!               T( I )= T( I ) - DOT_PRODUCT( RS( I,1:m ), X2 )
                T( I )= T( I ) - DDOT( m, RS( I, 1:m ), 1, X2, 1)
             ENDIF
           ENDDO

!
! initialize identity array to be used in DGETRS
!
           DO I= 1, k*2+m
             P1( I )= I
           ENDDO
!copy the factorization Li\Ui in LU, as appears after DGETRF applied to Mi,
           LU(:,:)=0
!copy of ( L1\U1 )
!        ( tempL2) in TEMP(1:k+m,k)

           call dcopy((k+m)*k,LURS(1),1,TEMP(1,1),1)
           LU(1:k,1:k)=TEMP(1:k,1:k)


           J=1
           DO I=k+1,k*2+m

             IF (P(I).LE.k+m) THEN
                 LU(I,1:k)=TEMP(k+J,:)
                 J=J+1
             ENDIF

           ENDDO


!copy (   U2  ) in LU(1:k*2+m,k+1:k+m)
!     ( L3\U3 )
!     (   L5  )
!

           call dcopy((k*2+m)*m, LURS((k+m)*k+1), 1, LU(1,k+1),1)

!copy (  U4   ) in  LU(k+1:k*2+m,k+m+1:k*2+m)
!     ( L6\U5 )
!
           DO I= 1,k
             call dcopy(k+m,LURS((k+m)*k+(k*2+m)*m+(I-1)*(k+m)+1),&
                        1,LU(k+1,k+m+I),1)
           ENDDO




!
! solve the system L*U x(1) = (T(p)-S(1p)*x(0)-R(2p)*x(2))
!
           CALL DGETRS( 'N', k*2+m, 1, LU(1,1), k*2+m, &
   &                    P1(1), T(1), k*2+m, INFO )

           RETURN
        END SUBROUTINE SOLVE_BLOCKG



        SUBROUTINE REDUCE_RHS( NRWBLK, P, F, T1, T2 )
!
!  ==== BABDCR package ================
!  Giuseppe Romanazzi, Pierluigi Amodio
!  Dipartimento di Matematica
!  Universita' di Bari
!  June 20, 2004
!  ====================================
!
! Purpose
! =======
!
!  REDUCE_RHS  reduces a  NRWBLK*2 vector
!
!     ( T(1) )
!     ( T(2) )
!
!  in a NRWBLK vector T' by means of the
!  factorization in REDUCE_BLOCK.
!  This reduction uses fill-in matrix F
!  and permutation matrix P that comes out from
!  the factorization of the matrix
!
!    ( S(1) R(1)      )
!    (      S(2) R(2) )
!
!  to obtain T' = T(q)-F*T(p),
!  where
!
!        P*( T(1) ) = ( T(p) )
!          ( T(2) )   ( T(q) )
!
!  On exit, REDUCE_RHS gives the computed vector T' and T(p)
!
!
! Parameters
! ==========
!
! Input variables
!   NRWBLK integer,              the dimension of T1 and T2, NRWBLK>0
!
!   P      integer,              NRWBLK*2 vector,
!                                permutation vector that comes out
!                                from REDUCE_BLOCK
!
!   F     double precision,      NRWBLK by NRWBLK array,
!                                fill-in matrix that comes out
!                                from REDUCE_BLOCK
!
!   T1    double precision,      NRWBLK vector,
!                                first block row of input vector to reduce
!
!   T2    double precision,      NRWBLK vector,
!                                second block row of input vector to reduce
!
! Output variables
!   T1    double precision,      NRWBLK vector,
!                                the vector T' obtained in the reduction
!
!   T2    double precision,      NRWBLK  vector,
!                                the vector T(p) obtained by the permutation
!
          INTEGER,  INTENT(IN)     :: NRWBLK, P( NRWBLK*2 )
          REAL(DP), INTENT(IN)     :: F( NRWBLK, NRWBLK )
          REAL(DP), INTENT(IN OUT) :: T1( NRWBLK ), T2( NRWBLK )

! Parameters
          REAL(DP), PARAMETER       :: ONE= 1.0D0, NONE= -1.0D0

! Local variables
          INTEGER :: I, Q( NRWBLK*2 )
          REAL(DP):: TEMP( NRWBLK*2 )

! Blas routine: DGEMV

! Intrinsic procedure(not used at the moment)
!          INTRINSIC MATMUL

!
!define the inverse of P
!
           DO I= 1, NRWBLK*2
             Q( P( I ) )= I
           ENDDO
!
! permute vector ( T(1) )
!                ( T(2) )
!
           TEMP( Q( 1:NRWBLK ) )= T1( 1:NRWBLK )
           TEMP( Q( NRWBLK+1:NRWBLK*2 ) )= T2( 1:NRWBLK )
!
! determine T(p)
!
           T2= TEMP( 1:NRWBLK )
!
! compute T' = T(q)-F*T(p)

           T1= TEMP( NRWBLK+1:NRWBLK*2 )
           CALL DGEMV( 'N', NRWBLK, NRWBLK, NONE, F, NRWBLK, T2, 1, &
                       ONE, T1, 1 )

! the same result may be obtained by means of the intrinsic lapack procedure MATMUL
!
!            T1= TEMP( NRWBLK+1:NRWBLK*2 )-MATMUL( F, T2 )

           RETURN
        END SUBROUTINE REDUCE_RHS

        SUBROUTINE SOLVE_BLOCK( NRWBLK, X0, X2, P, LURS, T )
!
!  ==== BABDCR package ================
!  Giuseppe Romanazzi, Pierluigi Amodio
!  Dipartimento di Matematica
!  Universita' di Bari
!  June 20, 2004
!  ====================================
!
! Purpose
! =======
!
!  SOLVE_BLOCK  computes the solution x(1) of a linear system
!
!          L*U x(1) = T(p) - S(1p)*x(0) - R(2p)*x(2)
!
!  (of dimension NRWBLK) obtained from the subroutines REDUCE_BLOCK
!  and REDUCE_RHS applied to the system
!
!   ( S(1) R(1)      ) * ( x(0) ) = ( T(1) )
!   (      S(2) R(2) )   ( x(1) )   ( T(2) )
!                        ( x(2) )
!
!  Each block is NRWBLK by NRWBLK.
!
!  On exit, SOLVE_BLOCK gives the solution x(1)
!
!
! Parameters
! ==========
!
! Input variables
!   NRWBLK   integer,            the dimension of blocks of the original matrix, NRWBLK>0
!
!   XO     double precision,     NRWBLK vector, solution x(0) previously obtained
!
!   X2     double precision,     NRWBLK vector, solution x(2) previously obtained
!
!   P      integer,              NRWBLK*2 vector,
!                                permutation vector that comes out from REDUCE_BLOCK
!
!   LURS   double precision,     NRWBLK*2 by NRWBLK array,
!                                RS array that comes out from REDUCE_BLOCK
!
!   T      double precision,     NRWBLK vector, the rhs T(p)
!
!
! Output variables
!   T      double precision,     NRWBLK vector, the solution x(1)
!

          INTEGER,  INTENT (IN)     ::  NRWBLK, P( NRWBLK*2 )
          REAL(DP), INTENT (IN)     ::  X0( NRWBLK ), X2( NRWBLK ), &
                                          LURS( NRWBLK*2, NRWBLK )
          REAL(DP), INTENT (IN OUT) ::  T( NRWBLK )

! Local variables
          INTEGER :: I, INFO, P1( NRWBLK )

! Lapack routine: DGETRS

! Blas routine: DDOT

! Intrinsic procedure (not used in this implementation)
!          INTRINSIC DOT_PRODUCT

!
! compute T(p) - S(1p)*x(0) - R(2p)*x(2)
!
           DO I= 1, NRWBLK
             IF ( P(I).LE.NRWBLK ) THEN
!               T( I )= T( I ) - DOT_PRODUCT( LURS( NRWBLK+I,1:NRWBLK ),X0 )
                T( I )= T( I ) - DDOT( NRWBLK, LURS( NRWBLK+I, 1:NRWBLK ), 1, X0, 1)
             ELSE
!               T( I )= T( I ) - DOT_PRODUCT( LURS( NRWBLK+I,1:NRWBLK ),X2 )
                T( I )= T( I ) - DDOT( NRWBLK, LURS( NRWBLK+I, 1:NRWBLK ), 1, X2, 1)
             ENDIF
           ENDDO
!
! initialize identity array to be used in DGETRS
!
           DO I= 1, NRWBLK
             P1( I )= I
           ENDDO
!
! solve system L*U x(1) = (T(p)-S(1p)*x(0)-R(2p)*x(2))
!
           CALL DGETRS( 'N', NRWBLK, 1, LURS(1,1), 2*NRWBLK, &
   &                    P1(1), T(1), NRWBLK, INFO )

           RETURN
        END SUBROUTINE SOLVE_BLOCK


   END SUBROUTINE PARABABDCR_SOLVG


!!!!!!!!!!!!!!!!!!!!!!
!!in this work reassemble is used even in main2.f90
          SUBROUTINE  REASSEMBLE( NRWBLK, NBLOCK, MATR )

            INTEGER,  INTENT(IN)     :: NRWBLK, NBLOCK
            REAL(DP), INTENT(IN OUT) :: MATR( NRWBLK, NRWBLK*NBLOCK )

! Local variables
            REAL(DP) :: TEMP( NRWBLK*2, NRWBLK*NBLOCK/2 )

! Intrinsic procedure
            INTRINSIC  RESHAPE


             IF ( NBLOCK.EQ.2 ) THEN
               TEMP( 1:NRWBLK, : )= MATR( :, 1:NRWBLK )
               TEMP( NRWBLK+1:NRWBLK*2, : )= MATR( :, NRWBLK+1:NRWBLK*2 )
               MATR= RESHAPE( TEMP, (/ NRWBLK, NRWBLK*2 /) )
             ELSE
               TEMP( 1:NRWBLK, 1:NRWBLK )= MATR( :, NRWBLK*3+1:NRWBLK*4 )
               TEMP( 1:NRWBLK, NRWBLK+1:NRWBLK*2 )= MATR( :, 1:NRWBLK )
               TEMP( NRWBLK+1:NRWBLK*2, : )= MATR( :, NRWBLK+1:NRWBLK*3 )
               MATR= RESHAPE( TEMP, (/ NRWBLK, NRWBLK*4 /) )
             ENDIF

             RETURN
          END SUBROUTINE REASSEMBLE

!
!  !if L is .TRUE.  then LOGTOINT is 1
!  !   L is .FALSE. then LOGTOINT is 0
!          FUNCTION LOGTOINT(L)
!           INTEGER::  LOGTOINT
!           LOGICAL::  L
!
!            IF (L) THEN
!              LOGTOINT=1
!            ELSE
!              LOGTOINT=0
!            ENDIF
!
!            RETURN
!         END FUNCTION LOGTOINT
!
!!function logdue:  computes the less grater integer of log_2(x)
FUNCTION LOGDUE(X)
   IMPLICIT NONE
   INTEGER             :: LOGDUE, Y
   INTEGER, INTENT(IN) :: X
    LOGDUE= 0
    Y=X
    DO WHILE ( Y.NE.1 )
     LOGDUE=LOGDUE+1
     IF (MOD(Y,2).NE.0) THEN
      !here Y/2 is  floor of real Y/2
      Y= Y/2 + 1
     ELSE
      Y=Y/2
     ENDIF
    ENDDO
    RETURN
END FUNCTION LOGDUE
!
!FUNCTION LENGTH( K, NFORPROC, NPROCS, NGSTEPS, NBLOKS )
!  IMPLICIT NONE
!  INTEGER :: LENGTH, L, M
!  INTEGER, INTENT(IN) :: K, NFORPROC, NPROCS, NGSTEPS, NBLOKS
!!  INTERFACE
!!    FUNCTION BASIC(K, NPROCS,NBLOKS,NFORPROC)
!!      INTEGER             :: BASIC
!!      INTEGER, INTENT(IN) :: K, NPROCS,NBLOKS,NFORPROC
!!    END FUNCTION BASIC
!!   END INTERFACE
!   IF (NPROCS.EQ.1) THEN
!     LENGTH=2*(NBLOKS+1) + 1
!   ELSE
!     L= 2**NFORPROC
!     IF ( K.EQ.0 ) THEN
!       LENGTH= 3 + 2*(L + NGSTEPS - NFORPROC)
!     ELSE
!      IF ( K.EQ.NPROCS-1 ) THEN
!        LENGTH= 2*(NBLOKS-L*(NPROCS-1))
!      ELSE
!       LENGTH=BASIC(K,NPROCS, NBLOKS, NFORPROC)
!       LENGTH= 2*(L + LENGTH)
!      ENDIF
!     ENDIF
!   ENDIF
!   RETURN
!END FUNCTION LENGTH
!
!
!
!!determines the number of  block [Si Ri] which are added to
!!initial blocks in MYMATR_A of proc with rank K in MPI_COMM_WORLD
!! k=1,...,NPROCS-1
!FUNCTION BASIC(K, NPROCS, NBLOKS, NFORPROC)
!  IMPLICIT NONE
!  INTEGER             :: BASIC, M, S
!  INTEGER, INTENT(IN) :: K, NPROCS, NBLOKS, NFORPROC
!  IF ( MOD(K,2).EQ.0 ) THEN
!    M= K/2
!    BASIC= 1
!    S=2
!    DO WHILE (   ( (MOD(M,2).EQ.0).AND.(NPROCS.GE.K+S+1) ).OR. &
!                 ( (MOD(M,2).EQ.0).AND.(NPROCS.LT.K+S+1).AND. &
!                 (NBLOKS.GT.(2**NFORPROC)*NPROCS) )   )
!      M= M/2
!      BASIC= BASIC+1
!      S=S*2
!    ENDDO
!  ELSE
!    BASIC= 0
!  ENDIF
!  RETURN
!END FUNCTION BASIC

!this function determines the number of blank blocks of size
!2*(m**2) in MYMATR_A before the blocks
!received from proc 0 of the initial matrix.
FUNCTION SPACE(P,NPROCS)
  USE PRECISION
   INTEGER, INTENT(IN) :: P, NPROCS
   INTEGER :: SPACE, L, K, H, H2, V(NPROCS)

     V(1)=0
     V(2)=1
     H=1
     H2=2
     DO L=2, LOGDUE(NPROCS)
       H=H*2
       H2=H2*2
       DO K= H+1, H2
         V(K)= L - V(H2-K+1)
       ENDDO
     ENDDO
     SPACE=V(P)
END FUNCTION SPACE


END MODULE PARABABDCR_PACKAGEG
