c$$$    !Paramod4.f90  - The Effective Cyclic Reduction Module
c$$$    !%%%%%%%%%%%%%%
c$$$    ! Compiling command
c$$$    !     mpif90 -c -u -mpilog  paramod4.f90
c$$$    !this create the object file 
c$$$    !    paramod4.o
c$$$    !%%%%%%%%%%%%%
      MODULE PRECISION
          INTEGER, PARAMETER :: DP=KIND(1.0D0)
      END MODULE PRECISION
    
      MODULE PARABABDCR_PACKAGE4
        USE PRECISION
c        USE MPI
      
        CONTAINS
      
         SUBROUTINE PARABABDCR_FACT4( K, LENGTHK, LENGTHVBK, NPROCS, &
                                      LOGNPROCS, NFORPROC, NRWBLK, &
                                      NBLOKS, MYMATR_A, &
                                      MYPERM, MYFILL_IN, MYINFO, OUT )
c$$$      !
c$$$      ! Parameters
c$$$      ! ==========
c$$$      !
c$$$      ! Input variables 
c$$$      !
c$$$      !   K  integer,                  rank of processor in current communication domain
c$$$      !                                K is in the set 0,...,NPROCS-1
c$$$      !
c$$$      !   LENGTHK integer,             size of third dimension of MYMATR_A, LENGTHK>0
c$$$      !
c$$$      !   NPROCS    integer,           number of processors, NPROCS>0
c$$$      !
c$$$      !   LOGNPROCS integer,           log2 of NPROCS
c$$$      !
c$$$      !   NFORPROC integer,            log2 of number of blocks [Si,Ri] in each proc
c$$$      !                                in the last proc there is at most 2^NFORPROC + 1 blocks
c$$$      !                                and at least 1 block, NFORPROC>0
c$$$      !
c$$$      !   NRWBLK integer,              the dimension of blocks S, R, Ba, Bb, NRWBLK>0
c$$$      !
c$$$      !   NBLOKS integer,              the number of blocks S(i) and R(i)
c$$$      !                                in the babd matrix, NBLOKS>0
c$$$      !
c$$$      !   MYMATR_A double precision,   NRWBLK by NRWBLK by LENGTHK array,
c$$$      !                                the blocks of babd matrix which are saved
c$$$      !                                in current processor
c$$$      !
c$$$      ! Output variables
c$$$      !   MYMATR_A double precision,   NRWBLK by NRWBLK by LENGTHK array,
c$$$      !                                part of the babd matrix factorization
c$$$      !
c$$$      !   MYPERM   integer,            if K>0,
c$$$      !                                 NRWBLK*2 by
c$$$      !                                 (LENGTHK-2)/2= (LENGTHK-2 - LOGTOINT(K.EQ.0) )/2 array,
c$$$      !                                 (since LOGTOINT(.TRUE.)=1, LOGTOINT(.FALSE.)=0)
c$$$      !                                 the permutations associated to the
c$$$      !                                 LU factorizations computed on proc K
c$$$      !                                if K=0,
c$$$      !                                 NRWBLK*2 by
c$$$      !                                 (LENGTHK-3)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0))/2 array,
c$$$      !                                 the permutations associated to the
c$$$      !                                 LU factorizations computed on proc 0, even
c$$$      !                                 associated to the factorization of the last
c$$$      !                                 2 by 2 block system
c$$$      !
c$$$      !   MYFILL_IN double precision,  if K>0, NRWBLK by NRWBLK by
c$$$      !                                 (LENGTHK-2)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
c$$$      !                                 the fill-in matrix generated in the reduction
c$$$      !                                 routine REDUCE_BLOCK called by proc K
c$$$      !                                if K=0, NRWBLK by NRWBLK by
c$$$      !                                 (LENGTHK-5)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
c$$$      !                                 the fill-in matrix generated in the reduction
c$$$      !                                 routine REDUCE_BLOCK called by proc 0
c$$$      !
c$$$      !   MYINFO   integer,
c$$$      !          = 0:  successful exit
c$$$      !          > 0:  if MYINFO = I, the I-th computed factorization
c$$$      !                (in proc K) is performed on a
c$$$      !                singular matrix; therefore the babd matrix is singular
c$$$      !
c$$$      !
c$$$      !  OUT  integer,      LOGNPROCS vector,
c$$$      !         OUT(I) = 0 : processor K communicates to one processor with idle greater than K,
c$$$      !                       at the I-th communication in the matrix reduction phase
c$$$      !         OUT(I) = 1 : processor K communicates to one processor with idle lesser than K,
c$$$      !                       at the I-th communication in the matrix reduction phase
c$$$      !
c$$$      
c$$$      
           INTEGER, INTENT(IN) :: K, LENGTHK, LENGTHVBK, NPROCS, 
                                  LOGNPROCS, NFORPROC, 
                                  NRWBLK, NBLOKS
c$$$      
c$$$      !dimension of MYMATR_A, MYFILL_IN, MYPERM changes through different processors
           REAL(DP), INTENT(IN OUT) :: MYMATR_A( NRWBLK, NRWBLK, &
                                       LENGTHK )      
           REAL(DP), INTENT(OUT)    :: MYFILL_IN( NRWBLK, NRWBLK, &
                                                (LENGTHK-5)/2 )
      
           INTEGER,  INTENT(OUT)    :: MYPERM( NRWBLK*2, &
                                      (LENGTHK-3)/2 ), &
                                       MYINFO, OUT( LOGNPROCS )
c$$$      
c$$$      ! Local variables in each processor
c$$$          INTEGER ::  MYHSTART, MYHEND, MYINDEXP, Z, I, MYLE, L, &
c$$$                      SOURCE, TAG, STATUS(MPI_STATUS_SIZE), &
c$$$                      EMERGENCYTAG, EMERGENCYMSG(2), &
c$$$                      IERR, JUMP, S, P, GSTEP, MSGTYPE, JUMP1, LSTART, &
c$$$                      EMPTY, PROD, event1a,event1b, & 
c$$$      !event2a,event2b,event3a,event3b,event4a,event4b, event5a,event5b&
c$$$                      event6a,event6b, MPE_Log_get_event_number, &
c$$$                      MPE_describe_state
c$$$          REAL(DP) :: SENDARRAY(NRWBLK,NRWBLK,2)
c$$$      
c$$$      ! Lapack routine : DGETRF
c$$$      ! Used subroutines: REASSEMBLE, REDUCE_BLOCK
c$$$      
c$$$      
c$$$      !     call MPE_INIT_LOG()
c$$$      
c$$$            event1a = MPE_Log_get_event_number()
c$$$            event1b = MPE_Log_get_event_number()
c$$$      !      event2a = MPE_Log_get_event_number()
c$$$      !      event2b = MPE_Log_get_event_number()
c$$$      !      event3a = MPE_Log_get_event_number()
c$$$      !      event3b = MPE_Log_get_event_number()
c$$$      !      event4a = MPE_Log_get_event_number()
c$$$      !      event4b = MPE_Log_get_event_number()
c$$$      !      event5a = MPE_Log_get_event_number()
c$$$      !      event5b = MPE_Log_get_event_number()
c$$$            event6a = MPE_Log_get_event_number()
c$$$            event6b = MPE_Log_get_event_number()
c$$$      !      event7a = MPE_Log_get_event_number()
c$$$      !      event7b = MPE_Log_get_event_number()
c$$$      
c$$$      
c$$$            ierr = MPE_Describe_state( event1a, event1b, &
c$$$                        "SendRecv MYMATR_A in reduction phase", "red" )
c$$$      !          ierr = MPE_Describe_state( event2a, event2b, &
c$$$      !                                     "Send EMERGMSG to proc NPROCS-1", "blue" )
c$$$      !          ierr = MPE_Describe_state( event3a, event3b, &
c$$$      !                                     "Receive MYMATR_A in reduction phase", "orange" )
c$$$      !          ierr = MPE_Describe_state( event4a, event4b, &
c$$$      !                                     "Receive  EMERGMSG ", "green" )
c$$$      !          ierr = MPE_Describe_state( event5a, event5b, &
c$$$      !                                    "Send MYMATR_A after receiving EMERGMSG", "cyan" )
c$$$             ierr = MPE_Describe_state( event6a, event6b, &
c$$$                        "Compute reduction - internal cycle", "yellow" )
c$$$      
c$$$      
c$$$            EMPTY=SPACE(K+1,NPROCS)
c$$$      
c$$$      !
c$$$      ! initialize MYINDEXP
c$$$      !
c$$$            MYINDEXP= 0
c$$$      
c$$$      !
c$$$      ! initialize MYLE, L
c$$$      !
c$$$      !  MYLE current number of blocks [Si Ri] to reduce
c$$$      !   in proc 0 the blocks Ba,Bb,0 are no considered
c$$$      !   then MYLE is 2**NFORPROC (or NBLOKS, if NPROCS.EQ.1) at beginning
c$$$      !   in proc 0,...NPROCS-2,
c$$$      !  L current index of last nonnull NRWBLK x NRWBLK block in MYMATR_A,
c$$$      !   in proc 0 the blocks Ba,Bb are not considered and even initial
c$$$      !   zero blocks
c$$$      !
c$$$      !     IF ( NPROCS.GT.1 ) THEN
c$$$      !       IF ( K.LT.NPROCS-1 ) THEN
c$$$               MYLE= 2**NFORPROC
c$$$      !       ELSE
c$$$      !         MYLE= LENGTHK/2
c$$$      !       ENDIF
c$$$      !     ELSE
c$$$      !       MYLE= NBLOKS
c$$$      !     ENDIF
c$$$      
c$$$           L= MYLE*2
c$$$      
c$$$            print *,'I''m proc ',K, &
c$$$                    ' inside _FACT my LENGTHK is ', LENGTHK ,&
c$$$                   'my MYLE is ',MYLE,' and NFORPROC is', NFORPROC
c$$$      !
c$$$      ! initialize JUMP, S
c$$$      !
c$$$      !  JUMP difference between index of block to reduce
c$$$      !  and index of used factorization in same array MYMATR_A
c$$$      !  S jump between processor wich are involved in communication
c$$$      !
c$$$           JUMP= 1
c$$$           S= 1
c$$$      
c$$$      
c$$$      
c$$$      !
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !!!!!!global reduction cycle!!!!!!!
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !
c$$$      ! divided in two part :
c$$$      ! 1. GSTEP=1,..,NFORPROC , in this part each processor work
c$$$      !      indipendently withou communication
c$$$      ! 2. GSTEP=NFORPROC+1,..,NFORPROC+LOGDUE(NPROCS),
c$$$      !      in this LOGDUE(NPROCS) steps each processor send and receive
c$$$      !      and after reduce its proper part
c$$$      
c$$$      !Part 1
c$$$           DO GSTEP=1,NFORPROC
c$$$      
c$$$             !initialize MYHSTART
c$$$      ! index of next block to reduce
c$$$               MYHSTART= 2*EMPTY + 1
c$$$      
c$$$               ierr = MPE_Log_event( event6a, 0, &
c$$$               "start compute-internal reduction")
c$$$      
c$$$               JUMP1= JUMP*2+1
c$$$      !!internal reduction cycle!!
c$$$               DO I= 1,MYLE/2-1
c$$$      !
c$$$      ! reshape [Ra Sb] as [Ra  and save in [ , ] columns by columns
c$$$      !                      Sb]
c$$$                 CALL REASSEMBLE( NRWBLK, 2, &
c$$$                                  MYMATR_A(:,:, &
c$$$                                  MYHSTART+JUMP:MYHSTART+JUMP+1) )
c$$$      !
c$$$      ! update Z, MYINDEXP
c$$$      !
c$$$                 Z= MYHSTART+JUMP1
c$$$                 MYINDEXP= MYINDEXP+1
c$$$      
c$$$            !     print *,'proc K', K, ' GSTEP', GSTEP,' I',I,'MYHSTART ',MYHSTART, &
c$$$            !             'JUMP ',JUMP, 'Z', Z
c$$$      
c$$$      !
c$$$      ! reduce blocks  [Sa Ra   ] in [S'a R'a]
c$$$      !                [   Sb Rb]
c$$$                 CALL REDUCE_BLOCK( NRWBLK, MYMATR_A(:,:,MYHSTART), &
c$$$                            MYMATR_A(:,:,&
c$$$                            MYHSTART+JUMP:MYHSTART+JUMP+1), &
c$$$                            MYMATR_A(:,:,Z), MYPERM(:,MYINDEXP), &
c$$$                            MYFILL_IN(:,:,MYINDEXP), MYINFO )
c$$$      
c$$$                 IF ( MYINFO.NE.0 ) THEN
c$$$                   MYINFO= MYINDEXP
c$$$                   RETURN
c$$$                 ENDIF
c$$$      !
c$$$      ! update MYHSTART
c$$$      !
c$$$                 MYHSTART= Z+1
c$$$               ENDDO
c$$$      
c$$$      !separating last step from DO cycle is necessary
c$$$      !only for processor NPROCS-1 when it has an odd number of blocks [Si Ri] to reduce
c$$$      
c$$$      !!last step of internal reduction cycle
c$$$      !
c$$$      ! reshape
c$$$      !
c$$$               CALL REASSEMBLE( NRWBLK, 2,&
c$$$                MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1) )
c$$$      !
c$$$      ! update Z, MYINDEXP
c$$$      !
c$$$                 Z= MYHSTART+JUMP1
c$$$                 MYINDEXP= MYINDEXP+1
c$$$      
c$$$      !           IF ( Z.GT.L ) THEN
c$$$      !             Z= L
c$$$      !           ENDIF
c$$$      
c$$$                ! print *,'proc K', K, ' GSTEP', GSTEP,' I',I,'MYHSTART ',MYHSTART, &
c$$$                !         'JUMP ',JUMP, 'Z', Z
c$$$      
c$$$                 CALL REDUCE_BLOCK( NRWBLK, MYMATR_A(:,:,MYHSTART), &
c$$$                          MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1), &
c$$$                          MYMATR_A(:,:,Z), MYPERM(:,MYINDEXP), &
c$$$                          MYFILL_IN(:,:,MYINDEXP), MYINFO )
c$$$      
c$$$                 IF ( MYINFO.NE.0 ) THEN
c$$$                    MYINFO= MYINDEXP
c$$$                    RETURN
c$$$                 ENDIF
c$$$      !!end internal reduction cycle!!
c$$$               ierr = MPE_Log_event( event6b, 0,&
c$$$                       "end compute-internal reduction")
c$$$      !
c$$$      ! update JUMP and MYLE
c$$$      !
c$$$                 JUMP= JUMP1
c$$$                 MYLE= MYLE-MYLE/2
c$$$      
c$$$      
c$$$            ENDDO
c$$$      
c$$$      
c$$$      !here L is even (2**NFORPROC)*2
c$$$            MYHSTART=2*EMPTY+1
c$$$            MYHEND= 2*EMPTY+L
c$$$            S=2
c$$$           ! LSTART=K
c$$$            PROD=MOD(K,S)
c$$$            JUMP=1
c$$$      
c$$$      
c$$$      !Part 2
c$$$           DO GSTEP= 1,LOGNPROCS
c$$$      
c$$$              SENDARRAY(:,:,1)=MYMATR_A(:,:,MYHSTART)
c$$$              SENDARRAY(:,:,2)=MYMATR_A(:,:,MYHEND)
c$$$      
c$$$              IF (PROD==0) THEN
c$$$                !communication to K+S/2
c$$$                !tagsend is 1, tagrecv is 2
c$$$                OUT(GSTEP)=0
c$$$                print *,'K ', K,' before SENDRECV to proc',&
c$$$                        K+ S/2,' with GSTEP', &
c$$$                GSTEP, 'PROD', PROD, 'S', S
c$$$      
c$$$                ierr = MPE_Log_event( event1a, 0, &
c$$$                 "start SENDRECV of MATR")
c$$$                CALL MPI_SENDRECV(SENDARRAY(1,1,1),2*NRWBLK*NRWBLK,MPI_DOUBLE_PRECISION, &
c$$$                                  K+ S/2, GSTEP, MYMATR_A(1,1,MYHEND+1), 2*NRWBLK*NRWBLK, &
c$$$                                  MPI_DOUBLE_PRECISION,K+ S/2,GSTEP,MPI_COMM_WORLD,STATUS, IERR )
c$$$                ierr = MPE_Log_event( event1b, 0, "end SENDRECV of MATR")
c$$$      
c$$$      
c$$$                MYHEND= MYHEND+2
c$$$                JUMP= L-1
c$$$              ELSE
c$$$                !communication to K-S/2
c$$$                OUT(GSTEP)=1
c$$$                 print *,'K ', K,' before SENDRECV to proc',K- S/2,'with GSTEP', GSTEP,&
c$$$                         'PROD', PROD, 'S', S
c$$$                ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MATR")
c$$$                CALL MPI_SENDRECV(SENDARRAY(1,1,1),2*NRWBLK*NRWBLK,MPI_DOUBLE_PRECISION, &
c$$$                                  K-S/2, GSTEP, MYMATR_A(1,1,MYHSTART-2), 2*NRWBLK*NRWBLK, &
c$$$                                  MPI_DOUBLE_PRECISION,K-S/2,GSTEP,MPI_COMM_WORLD,STATUS, IERR )
c$$$                ierr = MPE_Log_event( event1b, 0, "start SENDRECV of MATR")
c$$$      
c$$$      
c$$$                MYHSTART=MYHSTART-2
c$$$                JUMP=1
c$$$              ENDIF
c$$$      
c$$$      !update S
c$$$              S=S*2
c$$$      
c$$$      !update PROD
c$$$              PROD=1
c$$$              DO I=0,2**(GSTEP)-1
c$$$                PROD= PROD*MOD(K-I,S)
c$$$              ENDDO
c$$$      !        LSTART= LSTART - I
c$$$              ierr = MPE_Log_event( event6a, 0, "start compute-internal reduction")
c$$$      
c$$$      !
c$$$      ! reshape [Ra Sb] as [Ra  and save in [ , ] columns by columns
c$$$      !                      Sb]
c$$$                 CALL REASSEMBLE( NRWBLK, 2, &
c$$$                                  MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1) )
c$$$      
c$$$                 MYINDEXP= MYINDEXP+1
c$$$      
c$$$      !
c$$$      ! reduce blocks  [Sa Ra   ] in [S'a R'a]
c$$$      !                [   Sb Rb]
c$$$                 CALL REDUCE_BLOCK( NRWBLK, MYMATR_A(:,:,MYHSTART), &
c$$$                            MYMATR_A(:,:,MYHSTART+JUMP:MYHSTART+JUMP+1), &
c$$$                            MYMATR_A(:,:,MYHEND), MYPERM(:,MYINDEXP), &
c$$$                            MYFILL_IN(:,:,MYINDEXP), MYINFO )
c$$$      
c$$$                 IF ( MYINFO.NE.0 ) THEN
c$$$                   MYINFO= MYINDEXP
c$$$                   RETURN
c$$$                 ENDIF
c$$$      
c$$$                 ierr = MPE_Log_event( event6b, 0, "end compute-internal reduction")
c$$$      
c$$$                 L=L+2
c$$$      
c$$$           ENDDO
c$$$      
c$$$      !!factorize  the last 2by 2
c$$$      !fill the last block in MYMATR_A
c$$$           MYMATR_A(:,:,LENGTHK)= MYMATR_A(:,:,1)
c$$$      
c$$$      !!
c$$$      !! reshape the last blocks [R'''1 , Ba , Bb , S'''1]
c$$$      !! as a [S''1 R''1 ]
c$$$      !!      [ Ba   Bb  ]
c$$$      !!
c$$$          CALL REASSEMBLE( NRWBLK, 4, MYMATR_A( :,:,LENGTHK-3:LENGTHK ) )
c$$$      
c$$$      !!
c$$$      !! factorize last matrix
c$$$      !!
c$$$          MYINDEXP= MYINDEXP+1
c$$$          CALL DGETRF( NRWBLK*2, NRWBLK*2, MYMATR_A(1,1,LENGTHK-3), NRWBLK*2, &
c$$$                       MYPERM(1,(LENGTHK-3)/2), MYINFO )
c$$$          IF ( MYINFO.NE.0 ) THEN
c$$$            MYINFO= MYINDEXP
c$$$            STOP
c$$$          ENDIF
c$$$      
c$$$      
c$$$      !!
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !!!!!!!end global reduction cycle!!!!!!!
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !!
c$$$      !!
c$$$      
c$$$      
c$$$            RETURN
c$$$      
c$$$              CONTAINS
c$$$      
c$$$                SUBROUTINE REDUCE_BLOCK( NRWBLK, S, RS, R, P, F, INFO )
c$$$      
c$$$                  INTEGER,  INTENT (IN)     :: NRWBLK
c$$$                  REAL(DP), INTENT(IN OUT)  :: RS( NRWBLK*2, NRWBLK ), &
c$$$                                               S( NRWBLK, NRWBLK ), R( NRWBLK, NRWBLK )
c$$$                  REAL(DP), INTENT(OUT)     :: F( NRWBLK, NRWBLK )
c$$$                  INTEGER,  INTENT(OUT)     :: P( NRWBLK*2 ), INFO
c$$$      
c$$$      ! Parameters
c$$$                  REAL(DP), PARAMETER       :: ONE= 1.0D0, NONE= -1.0D0
c$$$      
c$$$      ! Local variables
c$$$                  INTEGER                   :: I, J, T, P1( NRWBLK )
c$$$                  REAL(DP)                  :: H( NRWBLK, NRWBLK ), L( NRWBLK, NRWBLK )
c$$$      
c$$$      ! Lapack routine : DGETRF ! Blas routines  : DTRSM, DGER
c$$$      
c$$$      
c$$$      !
c$$$      ! factorize RS by means of the LU factorization
c$$$      !
c$$$                   CALL DGETRF( NRWBLK*2, NRWBLK, RS(1,1), NRWBLK*2, P1(1), INFO )
c$$$                   IF ( INFO.NE.0 ) THEN
c$$$                     RETURN
c$$$                   ENDIF
c$$$      !
c$$$      ! compute fill-in block F = C*L^(-1)
c$$$      !
c$$$                   CALL DTRSM( 'R', 'L', 'N', 'U', NRWBLK, NRWBLK, ONE, RS(1,1), &
c$$$                               NRWBLK*2, RS(NRWBLK+1,1), NRWBLK*2 )
c$$$                   F= RS( NRWBLK+1:NRWBLK*2, : )
c$$$      !
c$$$      ! determine the permutation vector
c$$$      !
c$$$                   DO I= 1, NRWBLK*2
c$$$                     P( I )= I
c$$$                   ENDDO
c$$$                   DO I= 1, NRWBLK
c$$$                     J= P1( I )
c$$$                     IF ( J.NE.I ) THEN
c$$$                       T= P( I )
c$$$                       P( I )= P( J )
c$$$                       P( J )= T
c$$$                     ENDIF
c$$$                   ENDDO
c$$$      
c$$$      !
c$$$      ! compute
c$$$      ! S' = S(1q)-F*S(1p), R' = R(2q)-F*R(2p)
c$$$      ! in three steps:
c$$$      !
c$$$      ! first step,
c$$$      ! define H and L as blocks containing the nonnull rows of S(1q) and R(2q)
c$$$      !
c$$$                   H( :, : )= 0
c$$$                   L( :, : )= 0
c$$$                   DO I= 1, NRWBLK
c$$$                     J= P( I+NRWBLK )
c$$$                     IF ( J.LE.NRWBLK ) THEN
c$$$                       H( I, : )= S( J, : )
c$$$                     ELSE
c$$$                       L( I, : )= R( J-NRWBLK, : )
c$$$                     ENDIF
c$$$                   ENDDO
c$$$      !
c$$$      ! second step,
c$$$      ! compute  H = H-F*S(1p) and L = L-F*R(2p) by using
c$$$      ! knowledge of the nonnull rows of S(1p) and R(2p)
c$$$      !
c$$$                   DO I= 1, NRWBLK
c$$$                     IF ( P(I).LE.NRWBLK ) THEN
c$$$      ! nonnull row of S(1p)
c$$$                       RS( NRWBLK+I, : )= S( P(I), : )
c$$$                       CALL DGER( NRWBLK, NRWBLK, NONE, F(1,I), 1, S(P(I),1), &
c$$$                                  NRWBLK, H, NRWBLK )
c$$$                     ELSE
c$$$      ! nonnull row of R(2p)
c$$$                       RS( NRWBLK+I, : )= R( P(I)-NRWBLK, : )
c$$$                       CALL DGER( NRWBLK, NRWBLK, NONE, F(1,I), 1, R(P(I)-NRWBLK,1), &
c$$$                                  NRWBLK, L, NRWBLK )
c$$$                     ENDIF
c$$$                   ENDDO
c$$$      !
c$$$      ! third step,
c$$$      ! set S' = H and R' = L
c$$$      !
c$$$                   S= H
c$$$                   R= L
c$$$      
c$$$                   RETURN
c$$$                END SUBROUTINE REDUCE_BLOCK
c$$$      
c$$$            END SUBROUTINE PARABABDCR_FACT4
c$$$      
c$$$      
c$$$      
c$$$      
c$$$       SUBROUTINE  PARABABDCR_SOLV4( K, LENGTHK, LENGTHVBK, NPROCS, LOGNPROCS, NFORPROC, &
c$$$                                     NRWBLK, NBLOKS, MYMATR_A, &
c$$$                                     MYPERM, MYFILL_IN, OUT, MYVECTB )
c$$$      !
c$$$      ! Parameters
c$$$      ! ==========
c$$$      !
c$$$      ! Input variables
c$$$      !
c$$$      !   K  integer,                  rank of processor in current communication domain
c$$$      !                                K is in the set 0,...,NPROCS-1,
c$$$      !
c$$$      !   LENGTHK integer,             size of third dimension of MYMATR_A, LENGTHK>0
c$$$      !
c$$$      !   LENGTHVBK integer,           size of second dimension of MYVECTB, LENGTHVBK>0
c$$$      !                                for K=0 is NBLOKS+1
c$$$      !
c$$$      !   NPROCS    integer,           number of processors, NPROCS>0
c$$$      !
c$$$      !   LOGNPROCS integer,           log2 of NPROCS
c$$$      !
c$$$      !   NFORPROC integer,            log2 of number of blocks [Si,Ri] in each proc
c$$$      !                                in the last proc there is at most 2^NFORPROC + 1 blocks
c$$$      !                                and at least 1 block, NFORPROC>0
c$$$      !
c$$$      !   NRWBLK integer,              the dimension of blocks S, R, Ba, Bb, NRWBLK>0
c$$$      !
c$$$      !   NBLOKS integer,              the number of blocks S(i) and R(i)
c$$$      !                                in the babd matrix, NBLOKS>0
c$$$      !
c$$$      !   MYMATR_A double precision,   NRWBLK by NRWBLK by LENGTHK array,
c$$$      !                                part of the babd matrix factorization
c$$$      !
c$$$      !   MYPERM   integer,            if K>0,
c$$$      !                                 NRWBLK*2 by
c$$$      !                                 (LENGTHK-2)/2= (LENGTHK-2 - LOGTOINT(K.EQ.0) )/2 array,
c$$$      !                                 (since LOGTOINT(.TRUE.)=1, LOGTOINT(.FALSE.)=0)
c$$$      !                                 the permutations associated to the
c$$$      !                                 LU factorizations computed on proc K
c$$$      !                                if K=0,
c$$$      !                                 NRWBLK*2 by
c$$$      !                                 (LENGTHK-3)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0))/2 array,
c$$$      !                                 the permutations associated to the
c$$$      !                                 LU factorizations computed on proc 0, even
c$$$      !                                 associated to the factorization of the last
c$$$      !                                 2 by 2 block system
c$$$      !
c$$$      !   MYFILL_IN double precision,  if K>0, NRWBLK by NRWBLK by
c$$$      !                                 (LENGTHK-2)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
c$$$      !                                 the fill-in matrix generated in the reduction
c$$$      !                                 routine REDUCE_BLOCK called by proc K
c$$$      !                                if K=0, NRWBLK by NRWBLK by
c$$$      !                                 (LENGTHK-5)/2=(LENGTHK-2 - LOGTOINT(K.EQ.0)*3)/2 array,
c$$$      !                                 the fill-in matrix generated in the reduction
c$$$      !                                 routine REDUCE_BLOCK called by proc 0
c$$$      !
c$$$      !  OUT  integer,      LOGNPROCS vector,
c$$$      !         OUT(I) = 0 : processor K communicates to one processor with idle greater than K,
c$$$      !                       at the I-th communication in the rhs reduction phase
c$$$      !         OUT(I) = 1 : processor K communicates to one processor with idle lesser than K,
c$$$      !                       at the I-th communication in the rhs reduction phase
c$$$      !
c$$$      !
c$$$      !   MYVECTB double precision,     NRWBLK by LENGTHVBK array,
c$$$      !                                 the right hand side of the babd linear system
c$$$      !                                 stored in proc K
c$$$      !
c$$$      ! Output variables
c$$$      !   MYVECTB double precision,     NRWBLK by LENGTHVBK array,
c$$$      !                                 the solution of the linear system
c$$$      !
c$$$      !
c$$$      
c$$$      
c$$$          INTEGER, INTENT(IN) :: K, LENGTHK, LENGTHVBK, NPROCS, LOGNPROCS, NFORPROC, &
c$$$                                 NRWBLK, NBLOKS
c$$$      
c$$$      !dimension of MYMATR_A, MYFILL_IN, MYPERM, MYVECTB changes through different processors
c$$$      
c$$$      
c$$$          REAL(DP), INTENT(IN) :: MYMATR_A( NRWBLK, NRWBLK, LENGTHK ), &
c$$$                                  MYFILL_IN( NRWBLK, NRWBLK, (LENGTHK-5)/2  )
c$$$      
c$$$      
c$$$          INTEGER,  INTENT(IN) :: MYPERM( NRWBLK*2, (LENGTHK-3)/2), &
c$$$                                  OUT(LOGNPROCS)
c$$$      
c$$$          REAL(DP), INTENT(IN OUT) :: MYVECTB( NRWBLK, LENGTHVBK )
c$$$      
c$$$      
c$$$      
c$$$      ! Local variables in each processor
c$$$          INTEGER :: MYINDEXP, MYLE, L, MYHSTART, JUMP2, S, &
c$$$                     MYLEEQ2, MYHEND, GSTEP, INFO, LNEW, HNEW, &
c$$$                     TAG, STATUS(MPI_STATUS_SIZE), IERR, &
c$$$                     EMERGENCYTAG, EMERGENCYMSG(2), JUMP, P, DEST, &
c$$$                     Z, I,  SOURCE, HSTART, LOLD, HOLD, J, H2, &
c$$$                     H, EMPTY, event1a, event1b, event6a, event6b, &
c$$$                     event3a, event3b
c$$$      
c$$$          REAL(DP) :: MSG(NRWBLK), TC(NRWBLK*2), XMSG(NRWBLK*2)
c$$$      
c$$$      
c$$$            event1a = MPE_Log_get_event_number()
c$$$            event1b = MPE_Log_get_event_number()
c$$$            event3a = MPE_Log_get_event_number()
c$$$            event3b = MPE_Log_get_event_number()
c$$$            event6a = MPE_Log_get_event_number()
c$$$            event6b = MPE_Log_get_event_number()
c$$$      
c$$$      
c$$$           ierr = MPE_Describe_state( event1a, event1b, &
c$$$                                      "SendRecv MYVECTB in reduction phase", "green" )
c$$$           ierr = MPE_Describe_state( event6a, event6b, &
c$$$                                      "Compute rhs reduction - internal cycle", "brown" )
c$$$           ierr = MPE_Describe_state( event3a, event3b, &
c$$$                                      "Solve system", "blue" )
c$$$      
c$$$      
c$$$      
c$$$      
c$$$      ! Lapack routine : DGETRS
c$$$      ! Used subroutines: REDUCE_RHS, SOLVE_BLOCK
c$$$      
c$$$        EMPTY=SPACE(K+1,NPROCS)
c$$$      
c$$$      !
c$$$      ! initialize MYINDEXP
c$$$      !
c$$$           MYINDEXP= 0
c$$$      
c$$$      !
c$$$      ! initialize MYLE, L
c$$$      !
c$$$      !  MYLE current number of rhs  of size NRWBLK to reduce
c$$$      !   in proc 0, d(the rhs associated to boundary condition) is not considered
c$$$      !   then MYLE is 2**NFORPROC, if K<NPROCS-1 and MYLE is LENGTHVBK-1, if K=NPROCS-1
c$$$      !   (or NBLOKS, if NPROCS.EQ.1) at beginning
c$$$      !  L current index of last nonnull rhs of size NRWBLK in MYVECTB
c$$$      !   in proc 0,....,NPROCS-1, in proc 0 the rhs d is not considered
c$$$           MYLE= 2**NFORPROC
c$$$           L= MYLE
c$$$      
c$$$      !
c$$$      ! initialize JUMP2
c$$$      !
c$$$      !  JUMP2 is twice JUMP that is the difference between indeces of rhs to reduce
c$$$      !   in array MYVECTB
c$$$           JUMP2= 2
c$$$      
c$$$      
c$$$      !
c$$$      ! initialize JUMP
c$$$      !
c$$$      !  JUMP difference between index of block to reduce
c$$$      !  and index of used factorization in same array MYMATR_A
c$$$           JUMP= 1
c$$$      
c$$$      
c$$$      
c$$$      
c$$$      !
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !!!!!!global reduction cycle!!!!!!!
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !
c$$$      ! divided in two part :
c$$$      ! 1. GSTEP=1,..,NFORPROC , in this part each processor works
c$$$      !      indipendently without communication
c$$$      ! 2. GSTEP=NFORPROC+1,..,NFORPROC+LOGDUE(NPROCS),
c$$$      !      in this LOGDUE(NPROCS) steps each processor sends and
c$$$      !      receives and after reduces its proper vector
c$$$      
c$$$      !Part 1
c$$$           DO GSTEP=1,NFORPROC
c$$$      
c$$$            H= EMPTY + 1
c$$$          !  print *,'K ', K,' reduction rhs in GSTEP', GSTEP, &
c$$$          !          ' H',H,'JUMP', JUMP, 'NFORPROC',NFORPROC
c$$$      
c$$$      
c$$$            ierr = MPE_Log_event( event6a, 0, "start rhs compute-internal reduction")
c$$$      !
c$$$      !!!rhs internal reduction cycle!!!
c$$$      !
c$$$               DO I= 1,MYLE/2
c$$$                 MYINDEXP= MYINDEXP+1
c$$$      !
c$$$      ! reduce ( Ta ) in T'a
c$$$      !        ( Tb )
c$$$      !
c$$$      
c$$$      !           print *,'K ', K,' reduction rhs in GSTEP', GSTEP, &
c$$$      !                   ' H',H,'JUMP', JUMP, ' I',I
c$$$      
c$$$                 CALL REDUCE_RHS( NRWBLK, MYPERM(:,MYINDEXP), MYFILL_IN(:,:,MYINDEXP), &
c$$$                             MYVECTB(:,H), MYVECTB(:,H+JUMP) )
c$$$      !
c$$$      ! update H
c$$$      !
c$$$                 H= H+JUMP2
c$$$               ENDDO
c$$$      !
c$$$      !!!end rhs internal reduction cycle!!!
c$$$      !
c$$$               ierr = MPE_Log_event( event6b, 0, "end rhs compute-internal reduction")
c$$$      
c$$$               JUMP= JUMP2
c$$$               JUMP2= JUMP*2
c$$$               MYLE= MYLE- MYLE/2
c$$$      
c$$$            ENDDO
c$$$      
c$$$      !here L is  (2**NFORPROC)
c$$$            MYHSTART= EMPTY+1
c$$$            MYHEND= EMPTY+L
c$$$            S=2
c$$$      !here JUMP is the difference between the indeces of the
c$$$      !two right hand side to reduce
c$$$      
c$$$      ! print *,'K',K,'After part1 in reduction VECTB '
c$$$      !Part 2
c$$$           DO GSTEP= 1,LOGNPROCS
c$$$      
c$$$      
c$$$              IF (OUT(GSTEP)==0) THEN
c$$$                !communication to K+S/2
c$$$                !tagsend is 1, tagrecv is 2
c$$$      
c$$$                ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MYVECTB")
c$$$                CALL MPI_SENDRECV(MYVECTB(1,MYHSTART),NRWBLK, &
c$$$                                  MPI_DOUBLE_PRECISION, &
c$$$                                  K+ S/2, GSTEP, MYVECTB(1,MYHEND+1), NRWBLK, &
c$$$                                  MPI_DOUBLE_PRECISION,K+S/2,GSTEP ,MPI_COMM_WORLD,STATUS,IERR )
c$$$                ierr = MPE_Log_event( event1b, 0, "start SENDRECV of MYVECTB")
c$$$                MYHEND= MYHEND+1
c$$$                JUMP= L
c$$$              ELSE
c$$$                !communication to K-S/2
c$$$                ierr = MPE_Log_event( event1a, 0, "start SENDRECV of MYVECTB")
c$$$                CALL MPI_SENDRECV(MYVECTB(1,MYHSTART),NRWBLK,MPI_DOUBLE_PRECISION, &
c$$$                                  K-S/2, GSTEP, MYVECTB(1,MYHSTART-1), NRWBLK, &
c$$$                                  MPI_DOUBLE_PRECISION,K-S/2,GSTEP,MPI_COMM_WORLD,STATUS,IERR )
c$$$                ierr = MPE_Log_event( event1b, 0, "end SENDRECV of MYVECTB")
c$$$                MYHSTART=MYHSTART-1
c$$$                JUMP=1
c$$$              ENDIF
c$$$      
c$$$             ierr = MPE_Log_event( event6a, 0, "start rhs compute-internal reduction")
c$$$      
c$$$              MYINDEXP= MYINDEXP + 1
c$$$      !
c$$$      ! reduce ( Ta ) in T'a
c$$$      !        ( Tb )
c$$$      !
c$$$                 CALL REDUCE_RHS( NRWBLK, MYPERM(:,MYINDEXP), MYFILL_IN(:,:,MYINDEXP), &
c$$$                             MYVECTB(:,MYHSTART), MYVECTB(:,MYHSTART+JUMP) )
c$$$      
c$$$      
c$$$      !
c$$$      !!!end rhs internal reduction cycle!!!
c$$$      !
c$$$               ierr = MPE_Log_event( event6b, 0, "end rhs compute-internal reduction")
c$$$                 L=L+1
c$$$                 S=S*2
c$$$           ENDDO
c$$$      
c$$$      
c$$$      !
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !!!!!!end rhs global reduction cycle!!!!!!!
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !
c$$$      
c$$$      
c$$$      
c$$$      !
c$$$      !!
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !!!!!!!! 2 by 2 block linear system solution !!!!!!! to each processor
c$$$      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c$$$      !!
c$$$      !! solve 2 by 2 block system
c$$$      !!
c$$$      !!           ( S^(1) R^(1) ) (x(0)     )= ( T^(1) )
c$$$      !!           ( Ba      Bb  ) (x(NBLOKS))  (   d   )
c$$$      !
c$$$        !     IF (K.EQ.0) THEN
c$$$              TC( 1:NRWBLK*2 )= (/ MYVECTB(1:NRWBLK,1), MYVECTB(1:NRWBLK,LENGTHVBK) /)
c$$$              CALL DGETRS( 'N', NRWBLK*2, 1, MYMATR_A(1,1,LENGTHK-3), &
c$$$                           NRWBLK*2, MYPERM(1,(LENGTHK-3)/2), TC, NRWBLK*2, INFO )
c$$$              MYVECTB( :, 1 )= TC( 1:NRWBLK )
c$$$              MYVECTB( :, LENGTHVBK )= TC( NRWBLK+1:NRWBLK*2 )
c$$$      
c$$$      !        print *,'proc', K,'After solution of 2 by 2 MYVECTB', MYVECTB,' GSTEP=', GSTEP, 'MYLE',MYLE
c$$$        !   ENDIF
c$$$      
c$$$      
c$$$      
c$$$      !
c$$$      ! initialize L
c$$$      !  now L is the index of greater known x(l)
c$$$      !  that will be used in solution phase
c$$$            L= LENGTHVBK
c$$$            H= 1
c$$$            LNEW= L
c$$$            HNEW= H
c$$$      !
c$$$      !!!!solution part
c$$$      !!
c$$$      !1) first part I= LOGNPROCS,-1,1
c$$$      !
c$$$          ierr = MPE_Log_event( event3a, 0, "start solution system")
c$$$          DO GSTEP= LOGNPROCS,1,-1
c$$$      
c$$$      
c$$$           IF (OUT(GSTEP)==0) THEN
c$$$              J= L-1
c$$$              LNEW= J
c$$$            ELSE
c$$$              J= H+1
c$$$              HNEW= J
c$$$            ENDIF
c$$$      
c$$$            CALL SOLVE_BLOCK( NRWBLK, MYVECTB(:,H), &
c$$$                              MYVECTB(:,L), MYPERM(:,MYINDEXP), &
c$$$                              MYMATR_A(:,:,J*2-2:J*2-1), MYVECTB(:,J) )
c$$$            H=HNEW
c$$$            L=LNEW
c$$$            MYINDEXP=MYINDEXP-1
c$$$          ENDDO
c$$$      
c$$$      !
c$$$      !2) second part
c$$$      !
c$$$      ! the start value for H and L here are HNEW and LNEW
c$$$      
c$$$          JUMP= (2**NFORPROC)/2
c$$$          H2= 1
c$$$          DO GSTEP= NFORPROC,1,-1
c$$$      
c$$$            DO I=1, H2
c$$$              J=L-JUMP
c$$$      
c$$$              CALL SOLVE_BLOCK( NRWBLK, MYVECTB(:,H), &
c$$$                              MYVECTB(:,L), MYPERM(:,MYINDEXP), &
c$$$                              MYMATR_A(:,:,J*2-2:J*2-1), MYVECTB(:,J) )
c$$$              MYINDEXP=MYINDEXP-1
c$$$              L=H
c$$$              H=H-JUMP*2
c$$$            ENDDO
c$$$      
c$$$            L=LNEW
c$$$            H= L- JUMP
c$$$            JUMP=JUMP/2
c$$$            H2= H2*2
c$$$      
c$$$          ENDDO
c$$$      
c$$$      
c$$$      !!
c$$$      !!!
c$$$      !!!!!end solution part
c$$$      !!!
c$$$      !!
c$$$           ierr = MPE_Log_event( event3b, 0, "end solution system")
c$$$            RETURN
c$$$      
c$$$            CONTAINS
c$$$      
c$$$              SUBROUTINE REDUCE_RHS( NRWBLK, P, F, T1, T2 )
c$$$      !
c$$$      !  ==== BABDCR package ================
c$$$      !  Giuseppe Romanazzi, Pierluigi Amodio
c$$$      !  Dipartimento di Matematica
c$$$      !  Universita' di Bari
c$$$      !  June 20, 2004
c$$$      !  ====================================
c$$$      !
c$$$      ! Purpose
c$$$      ! =======
c$$$      !
c$$$      !  REDUCE_RHS  reduces a  NRWBLK*2 vector
c$$$      !
c$$$      !     ( T(1) )
c$$$      !     ( T(2) )
c$$$      !
c$$$      !  in a NRWBLK vector T' by means of the
c$$$      !  factorization in REDUCE_BLOCK.
c$$$      !  This reduction uses fill-in matrix F
c$$$      !  and permutation matrix P that comes out from
c$$$      !  the factorization of the matrix
c$$$      !
c$$$      !    ( S(1) R(1)      )
c$$$      !    (      S(2) R(2) )
c$$$      !
c$$$      !  to obtain T' = T(q)-F*T(p),
c$$$      !  where
c$$$      !
c$$$      !        P*( T(1) ) = ( T(p) )
c$$$      !          ( T(2) )   ( T(q) )
c$$$      !
c$$$      !  On exit, REDUCE_RHS gives the computed vector T' and T(p)
c$$$      !
c$$$      !
c$$$      ! Parameters
c$$$      ! ==========
c$$$      !
c$$$      ! Input variables
c$$$      !   NRWBLK integer,              the dimension of T1 and T2, NRWBLK>0
c$$$      !
c$$$      !   P      integer,              NRWBLK*2 vector,
c$$$      !                                permutation vector that comes out
c$$$      !                                from REDUCE_BLOCK
c$$$      !
c$$$      !   F     double precision,      NRWBLK by NRWBLK array,
c$$$      !                                fill-in matrix that comes out
c$$$      !                                from REDUCE_BLOCK
c$$$      !
c$$$      !   T1    double precision,      NRWBLK vector,
c$$$      !                                first block row of input vector to reduce
c$$$      !
c$$$      !   T2    double precision,      NRWBLK vector,
c$$$      !                                second block row of input vector to reduce
c$$$      !
c$$$      ! Output variables
c$$$      !   T1    double precision,      NRWBLK vector,
c$$$      !                                the vector T' obtained in the reduction
c$$$      !
c$$$      !   T2    double precision,      NRWBLK  vector,
c$$$      !                                the vector T(p) obtained by the permutation
c$$$      !
c$$$                INTEGER,  INTENT(IN)     :: NRWBLK, P( NRWBLK*2 )
c$$$                REAL(DP), INTENT(IN)     :: F( NRWBLK, NRWBLK )
c$$$                REAL(DP), INTENT(IN OUT) :: T1( NRWBLK ), T2( NRWBLK )
c$$$      
c$$$      ! Parameters
c$$$                REAL(DP), PARAMETER       :: ONE= 1.0D0, NONE= -1.0D0
c$$$      
c$$$      ! Local variables
c$$$                INTEGER :: I, Q( NRWBLK*2 )
c$$$                REAL(DP):: TEMP( NRWBLK*2 )
c$$$      
c$$$      ! Blas routine: DGEMV
c$$$      
c$$$      ! Intrinsic procedure(not used at the moment)
c$$$      !          INTRINSIC MATMUL
c$$$      
c$$$      !
c$$$      !define the inverse of P
c$$$      !
c$$$                 DO I= 1, NRWBLK*2
c$$$                   Q( P( I ) )= I
c$$$                 ENDDO
c$$$      !
c$$$      ! permute vector ( T(1) )
c$$$      !                ( T(2) )
c$$$      !
c$$$                 TEMP( Q( 1:NRWBLK ) )= T1( 1:NRWBLK )
c$$$                 TEMP( Q( NRWBLK+1:NRWBLK*2 ) )= T2( 1:NRWBLK )
c$$$      !
c$$$      ! determine T(p)
c$$$      !
c$$$                 T2= TEMP( 1:NRWBLK )
c$$$      !
c$$$      ! compute T' = T(q)-F*T(p)
c$$$      
c$$$                 T1= TEMP( NRWBLK+1:NRWBLK*2 )
c$$$                 CALL DGEMV( 'N', NRWBLK, NRWBLK, NONE, F, NRWBLK, T2, 1, &
c$$$                             ONE, T1, 1 )
c$$$      
c$$$      ! the same result may be obtained by means of the intrinsic lapack procedure MATMUL
c$$$      !
c$$$      !            T1= TEMP( NRWBLK+1:NRWBLK*2 )-MATMUL( F, T2 )
c$$$      
c$$$                 RETURN
c$$$              END SUBROUTINE REDUCE_RHS
c$$$      
c$$$              SUBROUTINE SOLVE_BLOCK( NRWBLK, X0, X2, P, LURS, T )
c$$$      !
c$$$      !  ==== BABDCR package ================
c$$$      !  Giuseppe Romanazzi, Pierluigi Amodio
c$$$      !  Dipartimento di Matematica
c$$$      !  Universita' di Bari
c$$$      !  June 20, 2004
c$$$      !  ====================================
c$$$      !
c$$$      ! Purpose
c$$$      ! =======
c$$$      !
c$$$      !  SOLVE_BLOCK  computes the solution x(1) of a linear system
c$$$      !
c$$$      !          L*U x(1) = T(p) - S(1p)*x(0) - R(2p)*x(2)
c$$$      !
c$$$      !  (of dimension NRWBLK) obtained from the subroutines REDUCE_BLOCK
c$$$      !  and REDUCE_RHS applied to the system
c$$$      !
c$$$      !   ( S(1) R(1)      ) * ( x(0) ) = ( T(1) )
c$$$      !   (      S(2) R(2) )   ( x(1) )   ( T(2) )
c$$$      !                        ( x(2) )
c$$$      !
c$$$      !  Each block is NRWBLK by NRWBLK.
c$$$      !
c$$$      !  On exit, SOLVE_BLOCK gives the solution x(1)
c$$$      !
c$$$      !
c$$$      ! Parameters
c$$$      ! ==========
c$$$      !
c$$$      ! Input variables
c$$$      !   NRWBLK   integer,            the dimension of blocks of the original matrix, NRWBLK>0
c$$$      !
c$$$      !   XO     double precision,     NRWBLK vector, solution x(0) previously obtained
c$$$      !
c$$$      !   X2     double precision,     NRWBLK vector, solution x(2) previously obtained
c$$$      !
c$$$      !   P      integer,              NRWBLK*2 vector,
c$$$      !                                permutation vector that comes out from REDUCE_BLOCK
c$$$      !
c$$$      !   LURS   double precision,     NRWBLK*2 by NRWBLK array,
c$$$      !                                RS array that comes out from REDUCE_BLOCK
c$$$      !
c$$$      !   T      double precision,     NRWBLK vector, the rhs T(p)
c$$$      !
c$$$      !
c$$$      ! Output variables
c$$$      !   T      double precision,     NRWBLK vector, the solution x(1)
c$$$      !
c$$$      
c$$$                INTEGER,  INTENT (IN)     ::  NRWBLK, P( NRWBLK*2 )
c$$$                REAL(DP), INTENT (IN)     ::  X0( NRWBLK ), X2( NRWBLK ), &
c$$$                                                LURS( NRWBLK*2, NRWBLK )
c$$$                REAL(DP), INTENT (IN OUT) ::  T( NRWBLK )
c$$$      
c$$$      ! Local variables
c$$$                INTEGER :: I, INFO, P1( NRWBLK )
c$$$      
c$$$      ! Lapack routine: DGETRS
c$$$      
c$$$      ! Blas routine: DDOT
c$$$      
c$$$      ! Intrinsic procedure (not used in this implementation)
c$$$      !          INTRINSIC DOT_PRODUCT
c$$$      
c$$$      !
c$$$      ! compute T(p) - S(1p)*x(0) - R(2p)*x(2)
c$$$      !
c$$$                 DO I= 1, NRWBLK
c$$$                   IF ( P(I).LE.NRWBLK ) THEN
c$$$      !               T( I )= T( I ) - DOT_PRODUCT( LURS( NRWBLK+I,1:NRWBLK ),X0 )
c$$$                      T( I )= T( I ) - DDOT( NRWBLK, LURS( NRWBLK+I, 1:NRWBLK ), 1, X0, 1)
c$$$                   ELSE
c$$$      !               T( I )= T( I ) - DOT_PRODUCT( LURS( NRWBLK+I,1:NRWBLK ),X2 )
c$$$                      T( I )= T( I ) - DDOT( NRWBLK, LURS( NRWBLK+I, 1:NRWBLK ), 1, X2, 1)
c$$$                   ENDIF
c$$$                 ENDDO
c$$$      !
c$$$      ! initialize identity array to be used in DGETRS
c$$$      !
c$$$                 DO I= 1, NRWBLK
c$$$                   P1( I )= I
c$$$                 ENDDO
c$$$      !
c$$$      ! solve system L*U x(1) = (T(p)-S(1p)*x(0)-R(2p)*x(2))
c$$$      !
c$$$                 CALL DGETRS( 'N', NRWBLK, 1, LURS(1,1), 2*NRWBLK, &
c$$$         &                    P1(1), T(1), NRWBLK, INFO )
c$$$      
c$$$                 RETURN
c$$$              END SUBROUTINE SOLVE_BLOCK
c$$$      
c$$$      
c$$$         END SUBROUTINE PARABABDCR_SOLV4
c$$$      
c$$$      
c$$$      !!!!!!!!!!!!!!!!!!!!!!
c$$$      !!in this work reassemble is used even in main2.f90
c$$$                SUBROUTINE  REASSEMBLE( NRWBLK, NBLOCK, MATR )
c$$$      
c$$$                  INTEGER,  INTENT(IN)     :: NRWBLK, NBLOCK
c$$$                  REAL(DP), INTENT(IN OUT) :: MATR( NRWBLK, NRWBLK*NBLOCK )
c$$$      
c$$$      ! Local variables
c$$$                  REAL(DP) :: TEMP( NRWBLK*2, NRWBLK*NBLOCK/2 )
c$$$      
c$$$      ! Intrinsic procedure
c$$$                  INTRINSIC  RESHAPE
c$$$      
c$$$      
c$$$                   IF ( NBLOCK.EQ.2 ) THEN
c$$$                     TEMP( 1:NRWBLK, : )= MATR( :, 1:NRWBLK )
c$$$                     TEMP( NRWBLK+1:NRWBLK*2, : )= MATR( :, NRWBLK+1:NRWBLK*2 )
c$$$                     MATR= RESHAPE( TEMP, (/ NRWBLK, NRWBLK*2 /) )
c$$$                   ELSE
c$$$                     TEMP( 1:NRWBLK, 1:NRWBLK )= MATR( :, NRWBLK*3+1:NRWBLK*4 )
c$$$                     TEMP( 1:NRWBLK, NRWBLK+1:NRWBLK*2 )= MATR( :, 1:NRWBLK )
c$$$                     TEMP( NRWBLK+1:NRWBLK*2, : )= MATR( :, NRWBLK+1:NRWBLK*3 )
c$$$                     MATR= RESHAPE( TEMP, (/ NRWBLK, NRWBLK*4 /) )
c$$$                   ENDIF
c$$$      
c$$$                   RETURN
c$$$                END SUBROUTINE REASSEMBLE
c$$$      
c$$$      !
c$$$      !  !if L is .TRUE.  then LOGTOINT is 1
c$$$      !  !   L is .FALSE. then LOGTOINT is 0
c$$$      !          FUNCTION LOGTOINT(L)
c$$$      !           INTEGER::  LOGTOINT
c$$$      !           LOGICAL::  L
c$$$      !
c$$$      !            IF (L) THEN
c$$$      !              LOGTOINT=1
c$$$      !            ELSE
c$$$      !              LOGTOINT=0
c$$$      !            ENDIF
c$$$      !
c$$$      !            RETURN
c$$$      !         END FUNCTION LOGTOINT
c$$$      !
c$$$      !!function logdue:  computes the less grater integer of log_2(x)
c$$$      FUNCTION LOGDUE(X)
c$$$         IMPLICIT NONE
c$$$         INTEGER             :: LOGDUE, Y
c$$$         INTEGER, INTENT(IN) :: X
c$$$          LOGDUE= 0
c$$$          Y=X
c$$$          DO WHILE ( Y.NE.1 )
c$$$           LOGDUE=LOGDUE+1
c$$$           IF (MOD(Y,2).NE.0) THEN
c$$$            !here Y/2 is  floor of real Y/2
c$$$            Y= Y/2 + 1
c$$$           ELSE
c$$$            Y=Y/2
c$$$           ENDIF
c$$$          ENDDO
c$$$          RETURN
c$$$      END FUNCTION LOGDUE
c$$$      !
c$$$      !FUNCTION LENGTH( K, NFORPROC, NPROCS, NGSTEPS, NBLOKS )
c$$$      !  IMPLICIT NONE
c$$$      !  INTEGER :: LENGTH, L, M
c$$$      !  INTEGER, INTENT(IN) :: K, NFORPROC, NPROCS, NGSTEPS, NBLOKS
c$$$      !!  INTERFACE
c$$$      !!    FUNCTION BASIC(K, NPROCS,NBLOKS,NFORPROC)
c$$$      !!      INTEGER             :: BASIC
c$$$      !!      INTEGER, INTENT(IN) :: K, NPROCS,NBLOKS,NFORPROC
c$$$      !!    END FUNCTION BASIC
c$$$      !!   END INTERFACE
c$$$      !   IF (NPROCS.EQ.1) THEN
c$$$      !     LENGTH=2*(NBLOKS+1) + 1
c$$$      !   ELSE
c$$$      !     L= 2**NFORPROC
c$$$      !     IF ( K.EQ.0 ) THEN
c$$$      !       LENGTH= 3 + 2*(L + NGSTEPS - NFORPROC)
c$$$      !     ELSE
c$$$      !      IF ( K.EQ.NPROCS-1 ) THEN
c$$$      !        LENGTH= 2*(NBLOKS-L*(NPROCS-1))
c$$$      !      ELSE
c$$$      !       LENGTH=BASIC(K,NPROCS, NBLOKS, NFORPROC)
c$$$      !       LENGTH= 2*(L + LENGTH)
c$$$      !      ENDIF
c$$$      !     ENDIF
c$$$      !   ENDIF
c$$$      !   RETURN
c$$$      !END FUNCTION LENGTH
c$$$      !
c$$$      !
c$$$      !
c$$$      !!determines the number of  block [Si Ri] which are added to
c$$$      !!initial blocks in MYMATR_A of proc with rank K in MPI_COMM_WORLD
c$$$      !! k=1,...,NPROCS-1
c$$$      !FUNCTION BASIC(K, NPROCS, NBLOKS, NFORPROC)
c$$$      !  IMPLICIT NONE
c$$$      !  INTEGER             :: BASIC, M, S
c$$$      !  INTEGER, INTENT(IN) :: K, NPROCS, NBLOKS, NFORPROC
c$$$      !  IF ( MOD(K,2).EQ.0 ) THEN
c$$$      !    M= K/2
c$$$      !    BASIC= 1
c$$$      !    S=2
c$$$      !    DO WHILE (   ( (MOD(M,2).EQ.0).AND.(NPROCS.GE.K+S+1) ).OR. &
c$$$      !                 ( (MOD(M,2).EQ.0).AND.(NPROCS.LT.K+S+1).AND. &
c$$$      !                 (NBLOKS.GT.(2**NFORPROC)*NPROCS) )   )
c$$$      !      M= M/2
c$$$      !      BASIC= BASIC+1
c$$$      !      S=S*2
c$$$      !    ENDDO
c$$$      !  ELSE
c$$$      !    BASIC= 0
c$$$      !  ENDIF
c$$$      !  RETURN
c$$$      !END FUNCTION BASIC
c$$$      
c$$$      !this function determines the number of blank blocks of size
c$$$      !2 x NRWBLK x NRWBLK  in MYMATR_A before the fill blocks
c$$$      !received from proc 0 of the initial matrix.
        FUNCTION SPACE(P,NPROCS)
          USE PRECISION
          INTEGER, INTENT(IN) :: P, NPROCS
          INTEGER :: SPACE, L, K, H, H2, V(NPROCS)
  
          V(1)=0
          V(2)=1
          H=1
          H2=2
          DO L=2, LOGDUE(NPROCS)
            H=H*2
            H2=H2*2
            DO K= H+1, H2
              V(K)= L - V(H2-K+1)
            ENDDO
          ENDDO
          SPACE=V(P)  
        END FUNCTION SPACE
  
  
      END MODULE PARABABDCR_PACKAGE4
  
  
  
